# log format constants
from typing import *
import contextlib
import copy
import getpass
import os

import numpy as np

from realhf.base.cluster import spec as cluster_spec

if TYPE_CHECKING:
    from realhf.api.core.config import ModelName
    from realhf.api.core.system_api import ModelShardID
    from realhf.base.topology import ParallelGrid, PipeModelDataParallelTopology


class GlobalMemoryBuffer:
    """Global buffer to avoid dynamic memory allocations.
    Caller should ensure that buffers of the same name
    are not used concurrently."""

    def __init__(self):
        self.buffer = {}

    def get_tensor(self, tensor_shape, dtype, name, force_zero: bool = False):
        import torch

        required_len = int(np.prod(tensor_shape))
        if self.buffer.get((name, dtype), None) is None:
            self.buffer[(name, dtype)] = torch.empty(
                required_len,
                dtype=dtype,
                device=torch.cuda.current_device(),
                requires_grad=False,
            )
        elif self.buffer[(name, dtype)].numel() < required_len:
            self.buffer[(name, dtype)] = torch.nn.functional.pad(
                self.buffer[(name, dtype)],
                (0, required_len - self.buffer[(name, dtype)].numel()),
                value=0,
            )
        res = self.buffer[(name, dtype)][0:required_len].view(*tensor_shape)
        if force_zero:
            res.zero_()
        return res


# constants in experiment instance scope
MODEL_SAVE_ROOT = f"{cluster_spec.fileroot}/checkpoints/{getpass.getuser()}"
LOG_ROOT = f"{cluster_spec.fileroot}/logs/{getpass.getuser()}"
RECOVER_ROOT = f"{cluster_spec.fileroot}/recover/{getpass.getuser()}"
SLURM_LOCK_FILE_NAME = f"{cluster_spec.fileroot}/logs/slurm_scheduler.lock"
PYTORCH_KERNEL_CACHE_PATH = (
    f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/torch/kernels"
)
TRITON_CACHE_PATH = f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/triton"
DATASET_CACHE_PATH = (
    f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/datasets"
)
PROFILER_CACHE_PATH = (
    f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/profiler"
)
TORCH_EXTENSIONS_DIR = (
    f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/torch/extensions"
)
QUICKSTART_EXPR_CACHE_PATH = (
    f"{cluster_spec.fileroot}/.cache/{getpass.getuser()}/"
)

# make directories if does not exist
os.makedirs(MODEL_SAVE_ROOT, exist_ok=True)
os.makedirs(LOG_ROOT, exist_ok=True)
os.makedirs(RECOVER_ROOT, exist_ok=True)
os.makedirs(PYTORCH_KERNEL_CACHE_PATH, exist_ok=True)
os.makedirs(TRITON_CACHE_PATH, exist_ok=True)
os.makedirs(DATASET_CACHE_PATH, exist_ok=True)
os.makedirs(PROFILER_CACHE_PATH, exist_ok=True)
os.makedirs(TORCH_EXTENSIONS_DIR, exist_ok=True)
os.makedirs(QUICKSTART_EXPR_CACHE_PATH, exist_ok=True)

# _model_name will be changed in the model_scope context manager
_model_name: "ModelName" = None

# constants in worker/process scope
_experiment_name = None
_trial_name = None

_grids: Dict["ModelName", "ParallelGrid"] = {}
_pgroups: Dict["ModelName", Any] = (
    {}
)  # torch.distributed.ProcessGroup, not type hint here to avoid importing torch
_pgroup_ranks: Dict["ModelName", List[int]] = {}
_self_group = None
_rank_mapping: Dict["ModelName", Dict["ModelShardID", int]] = {}
_global_memory_buffer: GlobalMemoryBuffer = GlobalMemoryBuffer()

# used only in scripts and tests
_fake_mp_world_size = None
_fake_mp_rank = None

# TODO: As in Megatron, we can set NCCL group options. Is it necessary?


@contextlib.contextmanager
def model_scope(model_name: "ModelName"):
    global _model_name
    assert _model_name is None
    _model_name = model_name
    yield
    assert _model_name == model_name
    _model_name = None


@contextlib.contextmanager
def model_scope_disabled():
    global _model_name
    assert _model_name is not None
    t, _model_name = _model_name, None
    yield
    _model_name = t


################# setter functions #################
def set_experiment_trial_names(expr_name: str, trial_name: str):
    global _experiment_name, _trial_name
    if _experiment_name is not None or _trial_name is not None:
        raise RuntimeError("Experiment and trial names are already set.")
    _experiment_name = expr_name
    _trial_name = trial_name


def set_grid(model_name: "ModelName", grid: "ParallelGrid"):
    global _grids
    if model_name in _grids:
        raise RuntimeError(f"Grid for model {model_name} is already set.")
    _grids[model_name] = grid


def set_parallelism_group(model_name: "ModelName", pgroup, ranks):
    global _pgroups
    if model_name in _pgroups:
        raise RuntimeError(
            f"Parallelism group for model {model_name} is already set."
        )
    _pgroups[model_name] = pgroup
    _pgroup_ranks[model_name] = ranks


def set_self_group(pgroup):
    global _self_group
    if _self_group is not None:
        raise RuntimeError("Self group is already set.")
    _self_group = pgroup


def set_rank_mapping(
    model_name: "ModelName",
    topo: "PipeModelDataParallelTopology",
    msid2mwid: Optional[Dict["ModelShardID", int]] = None,
):
    global _rank_mapping
    if model_name in _rank_mapping:
        raise RuntimeError(
            f"Rank mapping for model {model_name} is already set."
        )
    if msid2mwid is None:
        _rank_mapping[model_name] = {i: i for i in range(topo.world_size())}
    else:
        msid2mwid = {
            k: v for k, v in msid2mwid.items() if k.model_name == model_name
        }
        _rank_mapping[model_name] = {
            topo.get_rank(
                data=s.dp_rank, model=s.mp_rank, pipe=s.pp_rank
            ): mw_id
            for s, mw_id in msid2mwid.items()
        }


################# attribute functions #################
def use_te_impl() -> bool:
    try:
        import transformer_engine.pytorch as te

        TE_ENABLED = True
    except ImportError:
        TE_ENABLED = False
    return TE_ENABLED and os.getenv("REAL_LLM_USE_TE") == "1"


def sequence_parallel() -> bool:
    return grid().topology().sequence_parallel


def max_prompt_len() -> int:
    return grid().topology().max_prompt_len


def gradient_checkpointing() -> bool:
    return grid().topology().gradient_checkpointing


def has_model_name(name: str) -> bool:
    return name in _grids and _grids[name].global_rank != -1


def self_group():
    global _self_group
    assert _self_group is not None
    return _self_group


def model_name():
    if _model_name == None:
        raise RuntimeError(
            "Global constant `model_name` should be accessed in the `model_scope` context."
        )
    return _model_name


def experiment_name():
    if _experiment_name == None:
        raise RuntimeError(
            "Global constant `experiment_name` is accessed before set."
        )
    return _experiment_name


def trial_name():
    if _trial_name == None:
        raise RuntimeError(
            "Global constant `trial_name` is accessed before set."
        )
    return _trial_name


def grid() -> "ParallelGrid":
    if _model_name is None:
        raise RuntimeError(
            "Global constant `model_name` is accessed before set."
        )
    if _grids.get(_model_name, None) is None:
        raise RuntimeError(f"Grid for model {_model_name} is not set.")
    return _grids[_model_name]


def grid_of_model(model_name: str) -> "ParallelGrid":
    if _grids.get(model_name, None) is None:
        raise RuntimeError(f"Grid for model {model_name} is not set.")
    return _grids[model_name]


def parallelism_group():
    """Returns the 3D parallelism group of a specific model."""
    if _model_name is None:
        raise RuntimeError(
            "Global constant `model_name` is accessed before set."
        )
    if _pgroups.get(_model_name, None) is None:
        raise RuntimeError(
            f"Parallelism group for model {_model_name} is not set."
        )
    return _pgroups[_model_name]


def parallelism_group_ranks():
    if _model_name is None:
        raise RuntimeError(
            "Global constant `model_name` is accessed before set."
        )
    if _pgroup_ranks.get(_model_name, None) is None:
        raise RuntimeError(
            f"Parallelism group ranks for model {_model_name} is not set."
        )
    return _pgroup_ranks[_model_name]


def parallelism_group_size() -> int:
    """The 3D parallelism group size of a specific model, normally dp_size * pp_size * mp_size."""
    import torch.distributed as dist

    return dist.get_world_size(group=parallelism_group())


def parallelism_rank() -> int:
    """Return the rank of a specific model in its 3D parallelism group."""
    import torch.distributed as dist

    return dist.get_rank(group=parallelism_group())


def to_global_pg_rank(local_rank: int) -> int:
    global _rank_mapping
    if _rank_mapping is None or model_name() not in _rank_mapping:
        raise RuntimeError("Rank mapping is not set.")
    return _rank_mapping[model_name()][local_rank]


def rank_mapping_of_model(model_name: str) -> Dict["ModelShardID", int]:
    global _rank_mapping
    if _rank_mapping is None or _rank_mapping.get(model_name, None) is None:
        raise RuntimeError(f"Rank mapping for model {model_name} is not set.")
    return _rank_mapping[model_name]


def pipe_parallel_rank() -> int:
    return grid().get_pipe_parallel_rank()


def pipe_parallel_world_size() -> int:
    return grid().get_pipe_parallel_world_size()


def pipe_parallel_group():
    return grid().get_pipe_parallel_group()


def is_last_pipe_stage():
    return pipe_parallel_rank() == pipe_parallel_world_size() - 1


def is_first_pipe_stage():
    return pipe_parallel_rank() == 0


def next_pipe_stage():
    return (pipe_parallel_rank() + 1) % pipe_parallel_world_size()


def prev_pipe_stage():
    return (
        pipe_parallel_world_size() + pipe_parallel_rank() - 1
    ) % pipe_parallel_world_size()


def model_parallel_rank() -> int:
    try:
        return grid().get_tensor_model_parallel_rank()
    except RuntimeError as e:  # used only in scripts and tests
        if _fake_mp_rank is not None:
            return _fake_mp_rank
        else:
            raise e


def model_parallel_world_size() -> int:
    try:
        return grid().get_tensor_model_parallel_world_size()
    except RuntimeError as e:  # used only in scripts and tests
        if _fake_mp_world_size is not None:
            return _fake_mp_world_size
        else:
            raise e


def model_parallel_group():
    return grid().get_tensor_model_parallel_group()


def data_parallel_rank() -> int:
    return grid().get_data_parallel_rank()


def data_parallel_world_size() -> int:
    return grid().get_data_parallel_world_size()


def data_parallel_group():
    return grid().get_data_parallel_group()


def set_fake_mp_world_size(world_size):
    # used only in scripts and tests
    global _fake_mp_world_size
    _fake_mp_world_size = world_size


def set_fake_mp_rank(rank):
    # used only in scripts and tests
    global _fake_mp_rank
    _fake_mp_rank = rank


def set_fake_grid(model_name, rank, topo):
    # used only in scripts and tests
    from realhf.base.topology import FakeGrid

    global _grids
    _grids[model_name] = FakeGrid(rank=rank, topo=topo)


def get_global_memory_buffer():
    global _global_memory_buffer
    assert _global_memory_buffer is not None, "global memory buffer is not set"
    return _global_memory_buffer


def clear_global_memory_buffer():
    global _global_memory_buffer
    _global_memory_buffer = GlobalMemoryBuffer()



from typing import Union
import dataclasses


@dataclasses.dataclass
class PromptAnswerDatasetConfig:
    """Datasets used for SFT.

    The raw data must be a json or jsonl file, where each piece of data is a dictionary
    with keys `prompt` and `answer`. Both `prompt` and `answer` are strings.

    :param train_path: Path to the training dataset.
    :type train_path: str
    :param valid_path: Path to the evaluation dataset.
    :type valid_path: str
    :param max_seqlen: Maximum sequence length (prompt + answer).
        Sequences longer than this will be truncated.
    :type max_seqlen: int
    :param train_bs_n_seqs: Number of sequences in each batch during training.
    :type train_bs_n_seqs: int
    :param valid_bs_n_seqs: Number of sequences in each batch during validation.
    :type valid_bs_n_seqs: int
    """

    train_path: str = ""
    valid_path: str = ""
    max_seqlen: int = 1024
    train_bs_n_seqs: int = 256
    valid_bs_n_seqs: int = 256


@dataclasses.dataclass
class PairedComparisonDatasetConfig:
    """Datasets used for paired-comparison reward modeling and DPO.

    The raw data must be a json or jsonl file, where each piece of data is a dictionary
    with keys `prompt`, `pos_answers`, and `neg_answers`. `prompt` is a string.
    `pos_answers` and `neg_answers` are lists of strings. They must have the same size.

    Answer pairs of the same prompt will be sampled in the same batch. Hence, the number of sequences
    in each batch must be even, in the form of [P1A1+, P1A1-, P1A2+, P1A2-, P2A1+, P2A1-, P2A2+, P2A2-, ...],
    where `P` means prompt, `A` means answer, `+` means positive, and `-` means negative.

    The raw dataset may contain multiple answer pairs for each prompt. We will randomly sample
    `max_pairs_per_prompt` answer pairs for each prompt in each epoch.

    :param train_path: Path to the training dataset.
    :type train_path: str
    :param valid_path: Path to the evaluation dataset.
    :type valid_path: str
    :param max_pairs_per_prompt: Maximum number of answer pairs per prompt.
    :type max_pairs_per_prompt: int
    :param max_seqlen: Maximum sequence length (prompt + answer).
        Sequences longer than this will be truncated.
    :type max_seqlen: int
    :param train_bs_n_seqs: Number of sequences in each batch during training.
    :type train_bs_n_seqs: int
    """

    train_path: str = ""
    valid_path: str = ""
    max_pairs_per_prompt: int = 2
    max_seqlen: int = 1024
    train_bs_n_seqs: int = 256
    valid_bs_n_seqs: int = 256


@dataclasses.dataclass
class PromptOnlyDatasetConfig:
    """Datasets used for PPO RLHF.

    The raw data must be a json or jsonl file, where each piece of data is a dictionary
    with a single key called `prompt`, which is a string.

    :param path: Path to the training dataset.
    :type path: str
    :param max_prompt_len: Maximum prompt length.
        Prompts longer than this will be truncated.
    :type max_prompt_len: int
    :param train_bs_n_seqs: Number of prompts in each batch.
    :type train_bs_n_seqs: int
    :param pad_to_max_length: Whether to pad prompts to the maximum length.
    :type pad_to_max_length: bool
    """

    path: str = ""
    max_prompt_len: int = 256
    train_bs_n_seqs: int = 256
    pad_to_max_length: bool = False



# Copied from https://github.com/microsoft/DeepSpeed
from packaging.version import Version
import torch
import torch.distributed as dist

import realhf.base.constants as constants

ID_TO_DTYPE = [
    torch.float32,
    torch.float64,
    torch.complex64,
    torch.complex128,
    torch.float16,
    torch.bfloat16,
    torch.uint8,
    torch.int8,
    torch.int16,
    torch.int32,
    torch.int64,
    torch.bool,
]
DTYPE_TO_ID = {dtype: id_ for id_, dtype in enumerate(ID_TO_DTYPE)}


def _tensor_bytes(tensor):
    return tensor.numel() * tensor.element_size()


def can_send_recv() -> bool:
    # torch_version = Version(torch_info["version"])
    torch_version = Version(torch.__version__)
    sendrecv_min = Version("1.8")
    return torch_version >= sendrecv_min


assert can_send_recv()


def _is_valid_send_recv(src_stage, dest_stage):
    first_stage = 0
    last_stage = constants.grid().pipe_parallel_size - 1
    assert (
        abs(src_stage - dest_stage) == 1
        or (src_stage == first_stage and dest_stage == last_stage)
        or (src_stage == last_stage and dest_stage == first_stage)
    ), f"Functionality currently limited to send and receive between adjacent ranks only (src={src_stage}, dst={dest_stage})"


def send(tensor, dest_stage, async_op=False):
    # NOTE: The input is the stage id rather than the global rank
    src_stage = constants.grid().get_stage_id()
    _is_valid_send_recv(src_stage, dest_stage)

    dest_rank = constants.grid().stage_to_global(stage_id=dest_stage)
    send_method = dist.isend if async_op else dist.send
    return send_method(tensor, constants.to_global_pg_rank(dest_rank))


def recv(tensor, src_stage, async_op=False):
    # NOTE: The input is the stage id rather than the global rank
    dest_stage = constants.grid().get_stage_id()
    _is_valid_send_recv(src_stage, dest_stage)

    src_rank = constants.grid().stage_to_global(stage_id=src_stage)
    recv_method = dist.irecv if async_op else dist.recv
    return recv_method(tensor, constants.to_global_pg_rank(src_rank))



from typing import *

import numpy as np
import torch

from realhf.api.core import model_api
from realhf.base import constants, logging

logger = logging.getLogger("ReaL parallel")

# FIXME: gradients of layer norm layers (replicated for tensor parallel) may be incorrect
# keys used to identify modules
EMBEDDING_KEYS = [".wte", ".wpe"]  # dim=0 no bias
COLUMN_LINEAR_KEYS = [
    ".attn.c_attn.q_attn",
    ".attn.c_attn.k_attn",
    ".attn.c_attn.v_attn",
    ".mlp.c_fc",
    ".mlp.gate_proj",
    ".mlp.up_proj",
]  # dim=0 + partition bias
ROW_LINEAR_KEYS = [
    ".attn.c_proj",
    ".mlp.down_proj",
    ".mlp.c_proj",
]  # dim=1 + no partition bias

if constants.use_te_impl():
    COLUMN_LINEAR_KEYS = [
        ".attn.c_attn.q_attn",
        ".attn.c_attn.k_attn",
        ".attn.c_attn.v_attn",
        ".mlp.c_fc",
        ".mlp.fc1_weight",
    ]  # dim=0 + partition bias
    ROW_LINEAR_KEYS = [".attn.c_proj", ".mlp.fc2_weight"]


def tensor_slice_partition_fn(
    tensor: torch.Tensor,
    mp_rank: Optional[int],
    mp_world_size: int,
    dim: Optional[int],
) -> Union[List[torch.Tensor], torch.Tensor]:
    """Partition a pytorch tensor for model parallelism."""
    if dim is None:
        splits = [tensor for _ in range(mp_world_size)]
    else:
        assert tensor.shape[dim] % mp_world_size == 0
        splits = torch.split(
            tensor, tensor.shape[dim] // mp_world_size, dim=dim
        )
    if mp_rank is None:
        return [s.contiguous() for s in splits]
    else:
        return splits[mp_rank].contiguous()


def intervals_partition_fn(
    shape: torch.Size,
    mp_rank: Optional[int],
    mp_world_size: int,
    dim: Optional[int],
) -> Union[List[torch.Tensor], torch.Tensor]:
    """Get the intervals of a MP-partitioned tensor in the flatten view.

    For example, if a tensor of shape (2, 4) is partitioned along the second dimension
    into 2 parts, then the intervals are [(0, 2), (2, 4)].
    """
    assert mp_rank is not None
    param_size = int(np.prod(shape))
    if dim is None:
        return np.array([(0, param_size)], dtype=np.int64)

    if dim < 0:
        dim = len(shape) + dim
    assert shape[dim] % mp_world_size == 0

    if len(shape) == 1:
        assert dim == 0
        partition_size = shape[0] // mp_world_size
        return np.array(
            [(partition_size * mp_rank, partition_size * (mp_rank + 1))],
            dtype=np.int64,
        )
    else:
        assert len(shape) == 2, shape
        if dim == 0:
            row_start = mp_rank * shape[0] // mp_world_size
            row_end = (mp_rank + 1) * shape[0] // mp_world_size
            return np.array(
                [(row_start * shape[1], row_end * shape[1])], dtype=np.int64
            )
        else:
            assert dim == 1
            col_start = mp_rank * shape[1] // mp_world_size
            col_end = (mp_rank + 1) * shape[1] // mp_world_size
            return np.arange(shape[0], dtype=np.int64)[:, None] * shape[
                1
            ] + np.array([(col_start, col_end)], dtype=np.int64)


def shape_partition_fn(
    shape: torch.Size,
    mp_rank: Optional[int],
    mp_world_size: int,
    dim: Optional[int],
):
    """Get the partitioned shape of a tensor for model parallelism."""
    if dim is None:
        splits = [shape for _ in range(mp_world_size)]
    else:
        if dim < 0:
            dim = len(shape) + dim
        assert shape[dim] % mp_world_size == 0
        splits = [
            (*shape[:dim], shape[dim] // mp_world_size, *shape[dim + 1 :])
            for _ in range(mp_world_size)
        ]
    if mp_rank is None:
        return [s for s in splits]
    else:
        return splits[mp_rank]


def mp_partition_key(
    key: str,
    tensor_or_shape: torch.Tensor | torch.Size,
    mp_rank: Optional[int],
    mp_size: Optional[int],
    config: model_api.ReaLModelConfig,
    partition_fn: Callable[
        [torch.Tensor, Optional[int], int, Optional[int]],
        Union[List[torch.Tensor], torch.Tensor],
    ] = tensor_slice_partition_fn,
) -> torch.Tensor:
    """Partition a parameter of ReaLModel with name `key` by the method `partition_fn`."""
    if any([ek in key for ek in EMBEDDING_KEYS]):
        assert "weight" in key
        return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=0)
    elif key == f"{config.n_layers + 1}.weight":  # output head
        if (
            isinstance(tensor_or_shape, torch.Tensor)
            and tensor_or_shape.shape[0] == 1
        ) or (
            not isinstance(tensor_or_shape, torch.Tensor)
            and tensor_or_shape[0] == 1
        ):
            assert config.is_critic
            return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=None)
        else:
            return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=0)
    elif any([ck in key for ck in COLUMN_LINEAR_KEYS]):
        if (
            ("k_attn" in key) or ("v_attn" in key)
        ) and config.n_kv_heads % mp_size != 0:
            # logger.warning(f"Cannot split {config.n_kv_heads} kv heads evenly among "
            #                f"{mp_size} model parallel ranks, "
            #                f"use unsplitted linear for kv heads instead")
            return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=None)
        # partition both weight and bias
        return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=0)
    elif any([rk in key for rk in ROW_LINEAR_KEYS]):
        # only partition weight
        if "weight" in key:
            return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=1)
        else:
            assert "bias" in key, key
            return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=None)
    else:
        return partition_fn(tensor_or_shape, mp_rank, mp_size, dim=None)


def mp_partition_real_model_state_dict(
    state_dict: Dict[str, torch.Tensor],
    config: model_api.ReaLModelConfig,
    mp_size: int,
    mp_rank: Optional[int] = None,
) -> Union[Dict, List[Dict]]:
    if mp_size == 1:
        if mp_rank is None:
            return [state_dict]
        else:
            return state_dict

    new_state_dict = {}
    for k, v in state_dict.items():
        new_state_dict[k] = mp_partition_key(k, v, mp_rank, mp_size, config)

    if mp_rank is None:
        return [
            {k: v[mp_rank] for k, v in new_state_dict.items()}
            for mp_rank in range(mp_size)
        ]
    else:
        return new_state_dict


def get_real_model_param_shape(
    k: str,
    config: model_api.ReaLModelConfig,
    mp_size: int,
    sequence_parallel: bool,
) -> Tuple:
    if "wte.weight" in k:
        assert config.vocab_size % mp_size == 0
        return (config.vocab_size // mp_size, config.hidden_dim)
    elif "wpe.weight" in k:
        assert config.n_positions % mp_size == 0
        return (config.n_positions // mp_size, config.hidden_dim)
    elif ".ln." in k or ".ln_f." in k:
        return (config.hidden_dim,)
    elif k == f"{config.n_layers + 1}.weight":  # output head
        if config.is_critic and sequence_parallel:
            return (1, config.hidden_dim)
        elif not config.is_critic and mp_size > 1:
            assert config.vocab_size % mp_size == 0
            return (config.vocab_size // mp_size, config.hidden_dim)
        else:
            return (
                config.vocab_size if not config.is_critic else 1,
                config.hidden_dim,
            )
    elif any([ck in k for ck in COLUMN_LINEAR_KEYS]):
        if "k_attn" in k or "v_attn" in k:
            if "weight" in k:
                if config.n_kv_heads % mp_size == 0:
                    return (
                        config.head_dim * config.n_kv_heads // mp_size,
                        config.hidden_dim,
                    )
                else:
                    return (
                        config.head_dim * config.n_kv_heads,
                        config.hidden_dim,
                    )
            else:
                assert "bias" in k
                if config.n_kv_heads % mp_size == 0:
                    return (config.head_dim * config.n_kv_heads // mp_size,)
                else:
                    return (config.head_dim * config.n_kv_heads,)
        if "mlp" in k:
            if "weight" in k:
                return (config.intermediate_dim // mp_size, config.hidden_dim)
            else:
                assert "bias" in k
                return (config.intermediate_dim // mp_size,)
        if "weight" in k:
            assert config.hidden_dim // config.head_dim % mp_size == 0
            return (config.hidden_dim // mp_size, config.hidden_dim)
        else:
            assert "bias" in k
            return (config.hidden_dim // mp_size,)
    elif any([rk in k for rk in ROW_LINEAR_KEYS]):
        if "mlp" in k and "weight" in k:
            return (config.hidden_dim, config.intermediate_dim // mp_size)
        elif "attn" in k and "weight" in k:
            return (config.hidden_dim, config.hidden_dim // mp_size)
        elif "bias" in k:
            return (config.hidden_dim,)
        else:
            raise NotImplementedError(f"unkown shape of key {k}.")
    else:
        raise NotImplementedError(f"unkown shape of key {k}.")


def mp_merge_key(
    k: str,
    tensors: List[torch.Tensor],
    config: model_api.ReaLModelConfig,
) -> torch.Tensor:
    if any([ek in k for ek in EMBEDDING_KEYS]) and "weight" in k:
        return torch.cat(tensors, dim=0)
    elif k == f"{config.n_layers + 1}.weight" and not config.is_critic:
        return torch.cat(tensors, dim=0)
    elif any([ck in k for ck in COLUMN_LINEAR_KEYS]):
        return torch.cat(tensors, dim=0)
    elif any([rk in k for rk in ROW_LINEAR_KEYS]) and "weight" in k:
        return torch.cat(tensors, dim=1)
    else:
        return tensors[0]


def mp_merge_real_model_state_dict(
    state_dicts: List[Dict[str, torch.Tensor]],
    config: model_api.ReaLModelConfig,
) -> Dict:
    mp_size = len(state_dicts)
    if mp_size == 1:
        return state_dicts[0]

    new_state_dict = {}
    for k in state_dicts[0].keys():
        new_state_dict[k] = mp_merge_key(
            k, [sd[k] for sd in state_dicts], config
        )

    return new_state_dict


def partition_pipeline_layers(
    config: model_api.ReaLModelConfig,
    num_stages: int,
    embed_param_counter: Callable[[model_api.ReaLModelConfig], int],
    transformer_block_param_counter: Callable[
        [model_api.ReaLModelConfig, int], int
    ],
    head_param_counter: Callable[[model_api.ReaLModelConfig], int],
    method: str = "parameters_balanced",
) -> Dict[int, Tuple[int, int]]:
    from deepspeed.runtime import utils as ds_utils

    from realhf.base.datapack import (
        partition_balanced as true_partition_balanced,
    )

    # Each stage gets a simple uniform number of layers.
    param_counts = (
        [embed_param_counter(config)]
        + [
            transformer_block_param_counter(config, i)
            for i in range(config.n_layers)
        ]
        + [head_param_counter(config)]
    )
    parts = None
    if method == "uniform":
        parts = ds_utils.partition_uniform(
            num_items=config.n_layers + 2, num_parts=num_stages
        )
    elif method == "parameters":
        parts = ds_utils.partition_balanced(
            weights=param_counts, num_parts=num_stages
        )
    elif method == "parameters_balanced":
        param_counts = np.array(param_counts)
        parts = true_partition_balanced(nums=param_counts, k=num_stages)
    else:
        raise NotImplementedError(
            f"Partitioning method {method} not implemented."
        )

    stage_to_layer_idx = {}
    for stage in range(num_stages):
        start = parts[stage]
        stop = parts[stage + 1]
        stage_to_layer_idx[stage] = (start, stop)
    return stage_to_layer_idx


def pipeline_repartition_strategy(
    layer_mapping1: Dict[int, List[int]],
    layer_mapping2: Dict[int, List[int]],
):
    assert set(sum(layer_mapping1.values(), [])) == set(
        sum(layer_mapping2.values(), [])
    )
    assert all(isinstance(i, int) for i in layer_mapping1)
    assert all(isinstance(i, int) for i in layer_mapping2)

    layer_mapping1 = dict(sorted(layer_mapping1.items()))
    layer_mapping2 = dict(sorted(layer_mapping2.items()))

    layer_map: Dict[Tuple[int, int], List[int]] = {}
    for pp_rank2, layer_indices2 in layer_mapping2.items():
        for pp_rank1, layer_indices1 in layer_mapping1.items():
            layer_map[(pp_rank1, pp_rank2)] = sorted(
                list(set(layer_indices1).intersection(set(layer_indices2)))
            )

    return layer_map



from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union
import contextlib
import dataclasses
import functools
import json
import os

import torch
import torch.nn as nn
import torch.utils.checkpoint
import transformers

from realhf.api.core import model_api
from realhf.impl.model.modules import (
    CausalSelfAttentionLayer,
    LayerNormMLP,
    LlamaLayerNormMLP,
    LlamaRMSNorm,
)
from realhf.impl.model.parallelism.model_parallel.modules import (
    ColumnParallelLinear,
    parallel_lm_logits,
    ParallelEmbedding,
    RowParallelLinear,
)
from realhf.impl.model.utils.functional import compute_varlen_position_indices
import realhf.base.constants as constants
import realhf.base.logging as logging
import realhf.impl.model.parallelism.model_parallel.mappings as tensor_parallel

logger = logging.getLogger("ReaLModelBase")


@dataclasses.dataclass
class PipeTransferData:
    """Data structure for transferring data between stages.

    Each pipeline stage has exactly one PipeTransferData as the input and the output,
    no matter how many layers are in this stage.

    Attributes:
        pp_input: The input to the current stage. Usually hidden states
            with shape [bs, seq_len, hidden_dim].
        pp_output: The output of the current stage, also the input to the next stage.
            Usually hidden states with shape [bs, seq_len, hidden_dim].
        cu_seqlens: The cumulative sequence lengths of packed input_ids.
            Used by flash_attn_varlen_func. Will not be used during generation.
            It's configuration-like data that must be transfered from the first stage
            to the last. Shape [bs + 1].
        max_seqlen: The maximum sequence length of packed input_ids.
            Used by flash_attn_varlen_func. Will not be used during generation.
            It's configuration-like data that must be transfered from the first stage
            to the last.
        store_kv_cache: Whether to store the key and value cache for generation.
        attention_mask: The attention mask of the input, the same as huggingface transformers.
            Used by torch_attn_func to examine the outputs of PyTorch attention and flash
            attention are the same. Only for debugging. Shape [bs, seq_len].
    """

    pp_input: torch.Tensor = None
    pp_output: torch.Tensor = None

    # The followings are "configuration"-like data that should be passed across all stages.
    cu_seqlens: torch.Tensor = None
    max_seqlen: int = None
    store_kv_cache: bool = False


@dataclasses.dataclass
class PipeCacheData:
    """Data structure for caching data locally that will not be trasferred.

    Each layer has exactly one PipeCacheData as the input.
    If a pipeline stage has multiple layers, a list of PipeCacheData should be passed
    as the input. The cached tensors will be changed in-place.

    Attributes:
        input_ids: The input token ids. Used only at the first stage.
            Can be packed with shape [total_seq_len] or unpacked with shape [bs, seq].
        prompt_mask: Prompt mask used
        position_ids: Input position IDs. Can be resolved automatically in most cases.
            Used only at the first stage. The same shape as input_ids.
            If None, will be resolved automatically.
        k_cache: Key cache used for generation, shape [bs, max_seq, n_kv_heads, head_dim].
            Note that this is the cache for a specific layer, not for all layers.
        v_cache: Value cache used for generation, shape [bs, max_seq, n_kv_heads, head_dim].
            Note that this is the cache for a specific layer, not for all layers.
        cache_seqlens: The sequence lengths of the cached tokens. Used for generation. Shape [bs].
    """

    # Only cached in the first stage.
    packed_input_ids: torch.Tensor = None
    packed_position_ids: torch.Tensor = None
    # Cached in each transformer layer.
    k_cache: torch.Tensor = None
    v_cache: torch.Tensor = None
    cache_seqlens: torch.Tensor = None


class ReaLModelBlock(nn.Module):

    def __init__(
        self,
        config: model_api.ReaLModelConfig,
        layer_index: int,
        output_layernorm: bool = False,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        if dtype is None:
            dtype = torch.float16
        self.layer_index = layer_index
        self.attn = CausalSelfAttentionLayer(
            hidden_dim=config.hidden_dim,
            n_kv_heads=config.n_kv_heads,
            head_dim=config.head_dim,
            resid_pdrop=config.resid_pdrop,
            attn_pdrop=config.attn_pdrop,
            layer_index=layer_index,
            layer_norm_epsilon=config.layer_norm_epsilon,
            scale_attn_by_inverse_layer_idx=config.scale_attn_by_inverse_layer_idx,
            layer_norm_type=config.layer_norm_type,
            use_attention_bias=config.use_attention_bias,
            apply_rotary=config.apply_rotary,
            rotary_base=config.rotary_base,
            rotary_interleaved=config.rotary_interleaved,
            rotary_scaling=config.rotary_scaling,
            rotary_scaling_type=config.rotary_scaling_type,
            model_parallel=constants.model_parallel_world_size() > 1,
            gradient_accumulation_fusion=config.gradient_accumulation_fusion,
            dtype=dtype,
            device=device,
        )
        if config.mlp_type is None:
            self.mlp = LayerNormMLP(
                hidden_dim=config.hidden_dim,
                intermediate_dim=config.intermediate_dim,
                resid_pdrop=config.resid_pdrop,
                activation_function=config.activation_function,
                layer_norm_epsilon=config.layer_norm_epsilon,
                model_parallel=constants.model_parallel_world_size() > 1,
                gradient_accumulation_fusion=config.gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
        elif config.mlp_type == "llama":
            self.mlp = LlamaLayerNormMLP(
                hidden_dim=config.hidden_dim,
                intermediate_dim=config.intermediate_dim,
                activation_function=config.activation_function,
                layer_norm_epsilon=config.layer_norm_epsilon,
                model_parallel=constants.model_parallel_world_size() > 1,
                gradient_accumulation_fusion=config.gradient_accumulation_fusion,
                dtype=dtype,
                device=device,
            )
        self.output_layernorm = output_layernorm
        if output_layernorm:
            if config.layer_norm_type is None:
                layer_norm_fn = nn.LayerNorm
            elif config.layer_norm_type == "rms":
                layer_norm_fn = LlamaRMSNorm
            self.ln_f = layer_norm_fn(
                config.hidden_dim,
                eps=config.layer_norm_epsilon,
                dtype=dtype,
                device=device,
            )

    def forward(
        self, x: PipeTransferData, y: PipeCacheData
    ) -> PipeTransferData:
        pp_input = x.pp_input
        cu_seqlens = x.cu_seqlens
        k_cache = y.k_cache
        v_cache = y.v_cache
        cache_seqlens = y.cache_seqlens
        max_seqlen = x.max_seqlen
        if constants.gradient_checkpointing():
            pp_output, k, v = torch.utils.checkpoint.checkpoint(
                self._forward,
                pp_input,
                cu_seqlens,
                k_cache,
                v_cache,
                cache_seqlens,
                max_seqlen,
                use_reentrant=True,
            )
        else:
            pp_output, k, v = self._forward(
                pp_input,
                cu_seqlens,
                k_cache,
                v_cache,
                cache_seqlens,
                max_seqlen,
            )

        x.pp_output = pp_output
        if x.store_kv_cache:
            if y.k_cache is None:
                y.k_cache = k.detach()
            if y.v_cache is None:
                y.v_cache = v.detach()
            if y.cache_seqlens is None and x.cu_seqlens is not None:
                y.cache_seqlens = x.cu_seqlens[1:] - x.cu_seqlens[:-1]
        return x

    def _forward(
        self,
        pp_input: torch.Tensor,
        cu_seqlens: torch.Tensor,
        k_cache: Optional[torch.Tensor],
        v_cache: Optional[torch.Tensor],
        cache_seqlens: Optional[torch.Tensor],
        max_seqlen: int,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        h = pp_input
        attn_out, k, v = self.attn(
            hidden_states=h,
            cu_seqlens=cu_seqlens,
            max_seqlen=max_seqlen,
            k_cache=k_cache,
            v_cache=v_cache,
            cache_seqlens=cache_seqlens,
        )
        h = h + attn_out
        h = self.mlp(h) + h
        if self.output_layernorm:
            h = self.ln_f(h)
        return h, k, v


class VocabPositionEmbedding(nn.Module):

    def __init__(
        self,
        config: model_api.ReaLModelConfig,
        dtype: Optional[torch.dtype] = None,
        device: Optional[Union[str, torch.device]] = None,
    ):
        super().__init__()
        self.n_positions = config.n_positions

        model_parallel = constants.model_parallel_world_size() > 1
        if model_parallel:
            embed_cls = ParallelEmbedding
        else:
            embed_cls = nn.Embedding

        self.wte = embed_cls(
            config.vocab_size, config.hidden_dim, dtype=dtype, device=device
        )

        self.apply_abs_pos_embed = not config.apply_rotary
        if self.apply_abs_pos_embed:
            self.wpe = embed_cls(
                config.n_positions,
                config.hidden_dim,
                dtype=dtype,
                device=device,
            )

        self.embed_drop = nn.Dropout(config.embd_pdrop)

        self.self_attention_mask = torch.tril(
            torch.ones(
                (config.n_positions, config.n_positions),
                dtype=torch.bool,
                device=device,
            )
        )

    def forward(
        self, x: PipeTransferData, y: PipeCacheData
    ) -> PipeTransferData:
        # Set position ids.
        # if y.packed_position_ids is not None:
        #     raise ValueError("In our use cases, position_ids must be None.")
        y.packed_position_ids = compute_varlen_position_indices(
            total_seqlen=y.packed_input_ids.shape[0],
            cu_seqlens=x.cu_seqlens,
            seqlen_offsets=y.cache_seqlens,
        )
        if x.max_seqlen > self.n_positions:
            raise ValueError(
                f"max_seqlen ({x.max_seqlen}) must be <= n_positions ({self.n_positions})."
            )
        assert y.packed_position_ids.shape == y.packed_input_ids.shape, (
            y.packed_position_ids.shape,
            y.packed_input_ids.shape,
            x.cu_seqlens,
        )

        x.pp_output = self._forward(y.packed_input_ids, y.packed_position_ids)
        return x

    def _forward(
        self, input_ids: torch.LongTensor, position_ids: torch.LongTensor
    ) -> torch.Tensor:
        inputs_embeds = self.wte(input_ids)
        if self.apply_abs_pos_embed:
            inputs_embeds = inputs_embeds + self.wpe(position_ids)
        if constants.sequence_parallel():
            inputs_embeds = tensor_parallel.scatter_to_sequence_parallel_region(
                inputs_embeds
            )
            # `scatter_to_sequence_parallel_region` returns a view, which prevents
            # the original tensor from being garbage collected. Clone to facilitate GC.
            # Has a small runtime cost (~0.5%).
            inputs_embeds = inputs_embeds.clone()
            # with tensor_parallel.get_cuda_rng_tracker().fork():
            x = self.embed_drop(inputs_embeds)
        else:
            x = self.embed_drop(inputs_embeds)
        return x


class OutputHead(nn.Linear):

    def forward(
        self, x: PipeTransferData, y: PipeCacheData
    ) -> PipeTransferData:
        x.pp_output = nn.functional.linear(x.pp_input, self.weight, self.bias)
        return x

    def _forward(self, x: torch.Tensor):
        return super().forward(x)


class SequenceParallelCriticHead(nn.Linear):

    def forward(
        self, x: PipeTransferData, y: PipeCacheData
    ) -> PipeTransferData:
        all_gather_buffer = (
            tensor_parallel.gather_from_sequence_parallel_region(x.pp_input)
        )
        x.pp_output = nn.functional.linear(
            all_gather_buffer, self.weight, self.bias
        )
        return x

    def _forward(self, x: torch.Tensor):
        x = tensor_parallel.gather_from_sequence_parallel_region(x)
        return super().forward(x)


class SequenceParallelActorHead(ColumnParallelLinear):

    def forward(
        self, x: PipeTransferData, y: PipeCacheData
    ) -> PipeTransferData:
        x.pp_output = parallel_lm_logits(
            x.pp_input,
            self.weight,
            parallel_output=True,
            async_tensor_model_parallel_allreduce=self.async_tensor_model_parallel_allreduce,
            gradient_accumulation_fusion=self.gradient_accumulation_fusion,
            bias=self.bias,
        )
        # NOTE: the output is not the whole logits, but the logits for a part of tokens due to ColumnParallelLinear.
        # (However, data along the batch dim is all-gathered. No sequence parallel any more.)
        return x

    def _forward(self, x: torch.Tensor):
        return parallel_lm_logits(
            x,
            self.weight,
            parallel_output=True,
            async_tensor_model_parallel_allreduce=self.async_tensor_model_parallel_allreduce,
            gradient_accumulation_fusion=self.gradient_accumulation_fusion,
            bias=self.bias,
        )


# Paramter count, used for partitioning pipeline stages
# Ignoring tensor model parallel since it will evenly partition parameters.
def real_model_embed_param_count(config: model_api.ReaLModelConfig) -> int:
    count = config.vocab_size * config.hidden_dim
    if not config.apply_rotary:
        count += config.n_positions * config.hidden_dim
    return count


def real_model_embedding_param_keys(config: model_api.ReaLModelConfig) -> int:
    keys = ["0.wte.weight"]
    if not config.apply_rotary:
        keys += ["0.wpe.weight"]
    return keys


def real_model_tblock_param_count(
    config: model_api.ReaLModelConfig, idx: int
) -> int:
    count = 0
    nq = config.hidden_dim // config.head_dim

    if config.layer_norm_type is None:
        # nn.LayerNorm
        ln_count = 2 * config.hidden_dim
    elif config.layer_norm_type == "rms":
        # Llama RMSNorm
        ln_count = config.hidden_dim
    else:
        raise NotImplementedError()

    # layernorm qkv linear
    count += (
        ln_count
        + config.head_dim * (nq + config.n_kv_heads * 2) * config.hidden_dim
    )
    if config.use_attention_bias:
        count += config.head_dim * (nq + config.n_kv_heads * 2)

    # attention projection
    count += config.hidden_dim * config.hidden_dim
    if config.use_attention_bias:
        count += config.hidden_dim
    # NOTE: we ignore the parameters of RotoaryEmbedding here

    # mlp
    count += ln_count
    if config.mlp_type is None:
        count += 2 * config.hidden_dim * config.intermediate_dim
    elif config.mlp_type == "llama":
        count += 3 * config.hidden_dim * config.intermediate_dim
    else:
        raise NotImplementedError()

    if idx == config.n_layers - 1:
        count += ln_count
    return count


def real_model_tblock_param_keys(
    config: model_api.ReaLModelConfig, idx: int
) -> List[str]:
    keys = [
        f"{idx + 1}.attn.c_attn.ln.weight",
        f"{idx + 1}.attn.c_attn.q_attn.weight",
        f"{idx + 1}.attn.c_attn.k_attn.weight",
        f"{idx + 1}.attn.c_attn.v_attn.weight",
    ]
    if config.layer_norm_type is None:
        keys += [f"{idx + 1}.attn.c_attn.ln.bias"]
    if config.use_attention_bias:
        keys += [
            f"{idx + 1}.attn.c_attn.q_attn.bias",
            f"{idx + 1}.attn.c_attn.k_attn.bias",
            f"{idx + 1}.attn.c_attn.v_attn.bias",
        ]
    keys += [f"{idx + 1}.attn.c_proj.weight"]
    if config.use_attention_bias:
        keys += [f"{idx + 1}.attn.c_proj.bias"]
    keys += [f"{idx + 1}.mlp.ln.weight"]
    if config.layer_norm_type is None:
        keys += [f"{idx + 1}.mlp.ln.bias"]
    if config.mlp_type is None:
        keys += [
            f"{idx + 1}.mlp.c_fc.weight",
            f"{idx + 1}.mlp.c_proj.weight",
            f"{idx + 1}.mlp.c_fc.bias",
            f"{idx + 1}.mlp.c_proj.bias",
        ]
    elif config.mlp_type == "llama":
        keys += [
            f"{idx + 1}.mlp.gate_proj.weight",
            f"{idx + 1}.mlp.up_proj.weight",
            f"{idx + 1}.mlp.down_proj.weight",
        ]
    else:
        raise NotImplementedError()
    if idx == config.n_layers - 1:
        keys += [f"{idx + 1}.ln_f.weight"]
        if config.layer_norm_type is None:
            keys += [f"{idx + 1}.ln_f.bias"]
    return keys


def real_model_head_param_count(config: model_api.ReaLModelConfig) -> int:
    # NOTE: To hold consistent partitions between actor and critic models,
    # we count the number of parameters of the critic head as config.hidden_dim * config.vocab_size.
    # This is the intended behavior rather than a bug.
    return config.hidden_dim * config.vocab_size


def real_model_head_param_keys(config: model_api.ReaLModelConfig) -> List[str]:
    return [f"{config.n_layers + 1}.weight"]


def keys_from_layer_indices(
    config: model_api.ReaLModelConfig, layer_indices: List[int]
) -> List[str]:
    # assert _is_integer_list_contiguous(layer_indices)
    sd_keys = []
    for layer_idx in layer_indices:
        if layer_idx == 0:
            sd_keys += real_model_embedding_param_keys(config)
        elif layer_idx == config.n_layers + 1:
            sd_keys += real_model_head_param_keys(config)
        else:
            sd_keys += real_model_tblock_param_keys(config, layer_idx - 1)
    return sd_keys



from collections import defaultdict
from typing import Dict, List, Literal, Optional, Tuple
import fcntl
import re
import subprocess
import time

from realhf.base.cluster import spec as cluster_spec
from realhf.base.constants import SLURM_LOCK_FILE_NAME as LOCK_FILE_NAME
from realhf.scheduler.client import (
    JobException,
    JobInfo,
    JobState,
    SchedulerClient,
)
from realhf.scheduler.slurm.utils import (
    allocate_resources,
    SlurmLaunchInfo,
    SlurmResource,
    SlurmResourceNotEnoughException,
)
import realhf.base.logging as logging

logger = logging.getLogger("Slurm-scheduler")

SCHEDULING_RETRY_INTERVAL_SECONDS = 30
SCHEDULING_TIMEOUT_MAX_SECONDS = 3600 * 24


class SlurmSchedulerClient(SchedulerClient):
    """Uses Slurm (https://slurm.schedmd.com/overview.html)."""

    def __init__(self, expr_name, trial_name):
        super().__init__(expr_name, trial_name)

        self.__pending_jobs: Dict[str, SlurmLaunchInfo] = dict()
        self.__committed_jobs: Dict[str, SlurmLaunchInfo] = dict()

        self.__submission_counter = defaultdict(int)
        self.__wprocs_counter = defaultdict(int)

    def submit(self, worker_type, cmd, **kwargs):
        self.submit_array(worker_type, cmd, count=1, **kwargs)

    def submit_array(
        self,
        worker_type: str,
        cmd: str,  # XXX: should be None for workers
        count: int,
        cpu: int = 1,
        gpu_type: str = "geforce",
        gpu: int = 0,
        mem: int = 1024,  # MB
        env_vars: Optional[Dict] = None,
        container_image: str = cluster_spec.gpu_image,
        container_mounts: str = cluster_spec.default_mount,
        node_type: Optional[str] = None,
        nodelist: Optional[str] = None,
        exclude: Optional[str] = None,
        hostfile: bool = True,
        multiprog: bool = True,
        begin: str = None,
        deadline: str = None,
        time_limit: str = None,
    ):
        # record launch information, do not submit to slurm until `wait()` is called
        # NOTE: fractional GPU requirement will be resolved automatically in `__post_init__` of SlurnLaunchInfo
        launch_info = SlurmLaunchInfo(
            worker_type=worker_type,
            wprocs_in_job=count,
            resource_requirement=SlurmResource(
                mem=mem, cpu=cpu, gpu=gpu, gpu_type=gpu_type
            ),
            cmd=cmd,
            run_name=self.run_name,
            exper_name=self.expr_name,
            trial_name=self.trial_name,
            container_image=container_image,
            container_mounts=container_mounts,
            env_vars=env_vars,
            node_type=node_type,
            nodelist=nodelist,
            exclude=exclude,
            hostfile=hostfile,
            multiprog=multiprog,
            worker_submission_idx=self.__submission_counter[worker_type],
            begin=begin,
            deadline=deadline,
            time_limit=time_limit,
        )

        if (
            launch_info.slurm_name in self.__pending_jobs
            or launch_info.slurm_name in self.__committed_jobs
        ):
            raise ValueError(
                f"job name {launch_info.slurm_name} already existed."
            )

        if launch_info.multiprog:
            launch_info = self.__resolve_multiprog_file(launch_info)

        self.__submission_counter[worker_type] += 1
        self.__wprocs_counter[worker_type] += count

        self.__pending_jobs[launch_info.slurm_name] = launch_info
        logger.info(
            f"Registered Slurm job {launch_info.slurm_name} to scheduler."
        )

    def __resolve_multiprog_file(self, launch_info: SlurmLaunchInfo):
        worker_type = launch_info.worker_type
        cmd = launch_info.cmd.format(
            jobstep_id="%t",
            n_jobsteps=launch_info.n_jobsteps,
            worker_submission_index=self.__submission_counter[worker_type],
            wprocs_per_jobstep=launch_info.wprocs_per_jobstep,
            wprocs_in_job=launch_info.wprocs_in_job,
            wproc_offset=self.__wprocs_counter[worker_type],
        )
        launch_info.multiprog_content = (
            f"0-{launch_info.n_jobsteps - 1} {cmd}\n"
        )
        return launch_info

    def __allocate_and_commit_pending_jobs(self):
        """Allocate resources to all pending job specs.
        Generate hostfiles for each job info
        """
        start_time = time.monotonic()
        while True:
            try:
                fp = open(LOCK_FILE_NAME, "w")
                fcntl.flock(fp, fcntl.LOCK_EX)
                infos = list(self.__pending_jobs.values())
                infos = allocate_resources(infos)
                self.__pending_jobs = {info.slurm_name: info for info in infos}
                # logger.info("Allocated jobs: ")
                # for info in infos:
                #     logger.info(info)
                break
            except SlurmResourceNotEnoughException:
                logger.critical(
                    "Not enough resources to allocate all pending jobs. Retrying ..."
                )
                logger.warning(
                    "Time since start: %d seconds",
                    time.monotonic() - start_time,
                )
                fcntl.flock(fp, fcntl.LOCK_UN)
                time.sleep(SCHEDULING_RETRY_INTERVAL_SECONDS)
                if (
                    time.monotonic() - start_time
                    > SCHEDULING_TIMEOUT_MAX_SECONDS
                ):
                    raise TimeoutError(
                        f"Timeout waiting for {self.run_name} to schedule."
                    )

        try:
            for slurm_name, launch_info in self.__pending_jobs.items():
                launch_info.commit()
                self.__committed_jobs[slurm_name] = launch_info
            self.__pending_jobs = dict()
            states = [None for _ in self.__committed_jobs]
            while JobState.PENDING in states or None in states:
                time.sleep(0.1)
                states = self.__update_all()
            # time.sleep(2)
            fcntl.flock(fp, fcntl.LOCK_UN)
        except Exception as e:
            for launch_info in self.__committed_jobs.values():
                launch_info.cancel()
            fcntl.flock(fp, fcntl.LOCK_UN)
            raise e

    def stop(self, slurm_name: str):
        launch_info = self.__committed_jobs.get(slurm_name, None)
        if launch_info:
            launch_info.cancel()

    def stop_all(self, signal: Literal["SIGINT", "SIGKILL"] = "SIGKILL"):
        for launch_info in self.__committed_jobs.values():
            logger.info(f"Canceling job {launch_info.slurm_name}")
            launch_info.cancel(signal)
        time.sleep(0.2)
        # print("before stop wait", self.__pending_jobs)
        self.wait(
            check_status=(),
            remove_status=(
                JobState.CANCELLED,
                JobState.NOT_FOUND,
                JobState.FAILED,
                JobState.COMPLETED,
            ),
        )

    def find(self, slurm_name: str) -> JobInfo:
        launch_info = self.__committed_jobs.get(slurm_name, None)
        if launch_info is None or launch_info.job_info is None:
            return JobInfo(name=slurm_name, state=JobState.NOT_FOUND)
        else:
            return launch_info.job_info

    def find_all(self, job_name_regex: str = ".*") -> List[JobInfo]:
        self.__update_all()
        infos = []
        for r in self.__committed_jobs.values():
            if r.job_info is None:
                continue
            if re.fullmatch(job_name_regex, r.slurm_name):
                infos.append(r.job_info)
        return infos

    def wait(
        self,
        timeout=None,
        check_status: Tuple[JobState, ...] = (
            JobState.CANCELLED,
            JobState.FAILED,
            JobState.NOT_FOUND,
        ),
        remove_status: Tuple[JobState, ...] = (JobState.COMPLETED,),
        update=False,
    ):
        # before wait, commit all remaining pending jobs
        # TODO: grab global file lock to avoid multi-experiment deadlocks
        self.__allocate_and_commit_pending_jobs()
        # begin wait
        deadline = None if timeout is None else time.time() + timeout
        left = set(self.__committed_jobs.keys())
        num_jobs_left = len(left)
        logger.info(
            f"Waiting for {num_jobs_left} jobs. Jobs IDs: "
            f"{','.join(sorted([x.job_info.slurm_id for x in self.__committed_jobs.values()]))}."
        )
        while len(left) > 0:
            if len(left) < num_jobs_left:
                num_jobs_left = len(left)
                logger.info(f"Waiting for {num_jobs_left} jobs.")
            if deadline is not None and time.time() > deadline:
                raise TimeoutError(
                    f"Timeout waiting for {self.run_name}: {', '.join(sorted(left))}"
                )
            try:
                self.__update_all()
            except subprocess.CalledProcessError:
                logger.warning(
                    "Calling squeue failed. Check slurm manually if you continue to see this warning."
                )
                time.sleep(30)
                continue
            for job_slurm_name in list(left):
                launch_info = self.__committed_jobs[job_slurm_name]
                if launch_info.slurm_id is None:
                    continue
                if launch_info.job_info.state in check_status:
                    launch_info.show_log()
                    raise JobException(
                        run_name=self.run_name,
                        worker_type=launch_info.worker_type,
                        host=launch_info.job_info.host,
                        reason=launch_info.job_info.state,
                    )
                if launch_info.job_info.state in remove_status:
                    logger.info(
                        f"Job {launch_info.slurm_name} is {launch_info.job_info.state}.(Removed)"
                    )
                    left.remove(job_slurm_name)
                    if update:
                        self.__committed_jobs.pop(job_slurm_name)
            time.sleep(2)

    def __update_all(self):
        states = []
        for launch_info in self.__committed_jobs.values():
            state = launch_info.update()
            states.append(state)
        return states



from typing import Dict, List, Optional

import deepspeed
import torch
import torch.distributed as dist
import torch.utils.data
import tqdm

from realhf.base.namedarray import from_dict, NamedArray, recursive_apply
from realhf.impl.model.nn.real_llm_api import ReaLModel
from realhf.impl.model.nn.real_llm_generate import GenerationConfig
from realhf.impl.model.utils.functional import (
    build_shift_one_indices,
    gather_packed_shifted_log_probs,
)
import realhf.api.core.model_api as model_api
import realhf.base.constants as constants


def compute_packed_sft_loss(
    logits: torch.Tensor,
    packed_input_ids: torch.Tensor,
    cu_seqlens: torch.Tensor,
    prompt_mask: torch.Tensor,
    **kwargs,
) -> torch.Tensor:
    # **kwargs is used to ensure the correctness of invoking this function
    shift_one_indices = build_shift_one_indices(logits, cu_seqlens)
    logprobs = gather_packed_shifted_log_probs(
        logits, cu_seqlens, packed_input_ids
    ).float()
    prompt_mask = prompt_mask[shift_one_indices]
    logprobs = torch.where(prompt_mask, 0, logprobs)

    loss_sum = -logprobs.sum()

    with torch.no_grad():
        seqlogp = torch.zeros(
            cu_seqlens.shape[0] - 1, device=logits.device, dtype=torch.float64
        )
        for i in range(cu_seqlens.shape[0] - 1):
            m = prompt_mask[cu_seqlens[i] - i : cu_seqlens[i + 1] - i - 1]
            logp = logprobs[cu_seqlens[i] - i : cu_seqlens[i + 1] - i - 1]
            assert cu_seqlens[i + 1] - i - 1 <= logprobs.shape[0], (
                cu_seqlens,
                logprobs.shape,
            )
            seqlogp[i] = torch.where(m, 0.0, logp).sum() / (
                m.numel() - m.count_nonzero()
            )

    logging_ppl = (-seqlogp).exp().sum()
    token_denorm = prompt_mask.numel() - prompt_mask.count_nonzero()
    seq_denorm = torch.tensor(
        [cu_seqlens.shape[0] - 1], dtype=torch.float32, device=logits.device
    )

    # Logging loss and perplexity.
    logging_loss = loss_sum.detach().clone()
    logging_token_denorm = token_denorm.detach().clone().float()
    dist.all_reduce(
        logging_ppl, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )
    dist.all_reduce(
        logging_loss,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )
    dist.all_reduce(
        seq_denorm, op=dist.ReduceOp.SUM, group=constants.data_parallel_group()
    )
    dist.all_reduce(
        logging_token_denorm,
        op=dist.ReduceOp.SUM,
        group=constants.data_parallel_group(),
    )

    loss = loss_sum / token_denorm
    return loss, {
        "loss": logging_loss,
        "ppl": logging_ppl,
        "n_tokens": logging_token_denorm,
        "n_seqs": seq_denorm,
    }


class SFTInterface(model_api.ModelInterface):

    def train_step(self, model: model_api.Model, data: NamedArray) -> Dict:
        data = recursive_apply(data, lambda x: x.to(model.device))
        packed_input_ids: torch.Tensor = data[
            "packed_input_ids"
        ]  # shape [tot_seqlen]
        prompt_mask: torch.BoolTensor = data[
            "prompt_mask"
        ]  # shape [tot_seqlen]
        module: deepspeed.DeepSpeedEngine = model.module

        module.train()

        seqlens_cpu = data.metadata["seqlens"]
        max_seqlen = max(seqlens_cpu)
        cu_seqlens = torch.nn.functional.pad(
            torch.tensor(
                seqlens_cpu, dtype=torch.int32, device=model.device
            ).cumsum(0),
            (1, 0),
        )

        loss_fn_kwargs = dict(
            prompt_mask=prompt_mask,
            input_lens=cu_seqlens[1:]
            - cu_seqlens[
                :-1
            ],  # this is used to partition other loss_fn_kwargs into microbatches
        )
        stat = module.train_batch(
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
            loss_fn=compute_packed_sft_loss,
            num_micro_batches=constants.pipe_parallel_world_size() * 2,
            version_steps=model.version.global_step,
            **loss_fn_kwargs,
        )

        model.inc_version()

        res = dict()
        if stat:
            res = dict(
                loss=float(stat["loss"]) / int(stat["n_tokens"]),
                ppl=float(stat["ppl"]) / int(stat["n_seqs"]),
                n_tokens=int(stat["n_tokens"]),
                n_seqs=int(stat["n_seqs"]),
            )
        return res

    def save(self, model: model_api.Model, save_dir: str):
        module = model.module
        if not isinstance(module, ReaLModel):
            module = module.module
        module.save_to_hf(
            tokenizer=model.tokenizer,
            save_dir=save_dir,
        )

    @torch.no_grad()
    def evaluate(
        self,
        model_: model_api.Model,
        eval_dataloader: torch.utils.data.DataLoader,
    ) -> Dict:
        device = model_.device
        module = model_.module

        module.eval()
        losses = n_seqs = ppl = n_tokens = 0

        for step, data in enumerate(tqdm.tqdm(eval_dataloader)):
            seqlens_cpu = data.metadata["seqlens"]

            data = recursive_apply(data, lambda x: x.to(device))
            packed_input_ids: torch.Tensor = data[
                "packed_input_ids"
            ]  # shape [tot_seqlen]
            prompt_mask: torch.BoolTensor = data[
                "prompt_mask"
            ]  # shape [tot_seqlen]

            max_seqlen = max(seqlens_cpu)
            cu_seqlens = torch.nn.functional.pad(
                torch.tensor(
                    seqlens_cpu, dtype=torch.int32, device=model_.device
                ).cumsum(0),
                (1, 0),
            )

            loss_fn_kwargs = dict(
                prompt_mask=prompt_mask,
                input_lens=cu_seqlens[1:] - cu_seqlens[:-1],
            )
            stat = module.eval_batch(
                seqlens_cpu=seqlens_cpu,
                packed_input_ids=packed_input_ids,
                cu_seqlens=cu_seqlens,
                loss_fn=compute_packed_sft_loss,
                num_micro_batches=constants.pipe_parallel_world_size(),
                **loss_fn_kwargs,
            )

            if stat:
                losses += stat["loss"]
                n_tokens += stat["n_tokens"]
                n_seqs += stat["n_seqs"]
                ppl += stat["ppl"]

        res = dict()
        if stat:
            return dict(
                loss=float(losses / n_tokens),
                ppl=float(ppl / n_seqs),
                n_tokens=int(n_tokens),
                n_seqs=int(n_seqs),
            )
        return res

    @torch.no_grad()
    def inference(self, model: model_api.Model, data: NamedArray) -> Dict:
        device = model.device
        module = model.module
        module.eval()

        data = recursive_apply(data, lambda x: x.to(device))
        packed_input_ids: torch.Tensor = data["packed_input_ids"]
        cu_seqlens: torch.Tensor = data["cu_seqlens"]
        max_seqlen = int((cu_seqlens[1:] - cu_seqlens[:-1]).max())
        seqlens_cpu = data.metadata["seqlens"]

        logits = module.forward(
            seqlens_cpu=seqlens_cpu,
            packed_input_ids=packed_input_ids,
            cu_seqlens=cu_seqlens,
        )
        x = from_dict(dict(logits=logits))
        x.register_metadata(**data.medata)
        return x

    # for testing only
    @torch.no_grad()
    def generate(
        self,
        model: model_api.Model,
        data: NamedArray,
        gconfig: GenerationConfig,
    ) -> NamedArray:
        module = model.module

        module.eval()

        data = recursive_apply(data, lambda x: x.to(model.device))

        res = module.generate(
            seqlens_cpu=data.metadata["seqlens"],
            tokenizer=model.tokenizer,
            packed_input_ids=data["packed_input_ids"],
            cu_seqlens=data["cu_seqlens"],
            gconfig=gconfig,
        )
        if res is None:
            return dict()

        gen_tokens, logprobs, logits_mask, *_ = res

        return dict(
            gen_tokens=gen_tokens,
            log_probs=logprobs,
            logits_mask=logits_mask,
        )


model_api.register_interface("sft", SFTInterface)



from collections import defaultdict
from typing import *
import dataclasses
import itertools

import torch.distributed

from realhf.api.core import system_api
from realhf.api.core.config import ModelName
from realhf.base import topology
from realhf.impl.model.comm.global_comm import filter_match_mwids
from realhf.impl.model.comm.param_realloc import pipeline_repartition_strategy


@dataclasses.dataclass(unsafe_hash=True)
class DataTransferPair:
    src: ModelName
    src_dp_rank: int
    dst: ModelName
    dst_dp_rank: int


@dataclasses.dataclass
class DataTransferInfo:
    # Groups for data transfer among model workers.
    data_transfer_groups: Dict[DataTransferPair, torch.distributed.ProcessGroup]
    data_transfer_src_ranks: Dict[DataTransferPair, int]
    data_transfer_dst_ranks: Dict[DataTransferPair, List[int]]


def setup_data_transfer(
    model_topos: Optional[
        Dict[str, topology.PipeModelDataParallelTopology]
    ] = None,
    msid2mwid: Optional[Dict[system_api.ModelShardID, int]] = None,
    data_transfer_pairs: Optional[List[Tuple[ModelName, ModelName]]] = None,
) -> DataTransferInfo:

    mw_dp_ranks: Dict[Tuple[ModelName, int], List[int]] = {}
    mw_dp_head_ranks: Dict[ModelName, List[int]] = {}
    if model_topos is not None:
        assert msid2mwid is not None
        for model_name, topo in model_topos.items():
            mw_dp_head_ranks[model_name] = filter_match_mwids(
                model_name,
                topo,
                msid2mwid,
                pipe=topo.get_dim("pipe") - 1,
                model=0,
            )
            dp_size = topo.get_dim("data")
            for dp_i in range(dp_size):
                mw_dp_ranks[model_name, dp_i] = filter_match_mwids(
                    model_name,
                    topo,
                    msid2mwid,
                    data=dp_i,
                )

    data_transfer_groups, data_transfer_src_ranks = {}, {}
    data_transfer_dst_ranks = {}
    if data_transfer_pairs is not None:
        for src, dst in data_transfer_pairs:
            src_topo = model_topos[src]
            dst_topo = model_topos[dst]
            for src_dp, dst_dp in itertools.product(
                range(src_topo.get_dim("data")), range(dst_topo.get_dim("data"))
            ):
                key = DataTransferPair(
                    src=src, src_dp_rank=src_dp, dst=dst, dst_dp_rank=dst_dp
                )
                src_mw_rank = mw_dp_head_ranks[src][src_dp]
                dst_mw_ranks = mw_dp_ranks[dst, dst_dp]
                data_transfer_dst_ranks[key] = dst_mw_ranks
                if src_mw_rank not in dst_mw_ranks:
                    _ranks = [src_mw_rank] + dst_mw_ranks
                else:
                    _ranks = dst_mw_ranks
                data_transfer_groups[key] = topology.new_or_get_group(
                    _ranks, backend="nccl"
                )
                data_transfer_src_ranks[key] = src_mw_rank

    return DataTransferInfo(
        data_transfer_groups=data_transfer_groups,
        data_transfer_src_ranks=data_transfer_src_ranks,
        data_transfer_dst_ranks=data_transfer_dst_ranks,
    )


@dataclasses.dataclass
class DataTransferSenderStep:
    rank: int
    dst_ranks: List[int]
    group: torch.distributed.ProcessGroup
    key: str
    buf_indices: List[int]
    seqlens: List[int]


@dataclasses.dataclass
class DataTransferReceiverStep:
    rank: int
    src: int
    dst_ranks: List[int]
    group: torch.distributed.ProcessGroup
    key: str
    buf_indices: List[int]
    seqlens: List[int]


def derive_data_transfer_plan(
    keys: List[str],
    global_buffer_indices: List[int],
    global_seqlens: List[int],
    consumer_name: ModelName,
    consumer_mapping: Dict[int, List[int]],
    producer_names: Dict[str, ModelName],
    producer_mappings: Dict[Tuple[ModelName, str], Dict[int, List[int]]],
    data_transfer_info: DataTransferInfo,
) -> List[DataTransferReceiverStep | DataTransferSenderStep]:
    comm_plan = []

    for k in keys:
        producer_name = producer_names[k]
        producer_mapping = producer_mappings[(producer_name, k)]

        # partition mapping starts from zero, which is different from buffer indices
        repart_strat = pipeline_repartition_strategy(
            producer_mapping, consumer_mapping
        )

        for (dp_i, dp_j), comm_slots in repart_strat.items():
            if len(comm_slots) == 0:
                continue

            group_key = DataTransferPair(
                src=producer_name,
                src_dp_rank=dp_i,
                dst=consumer_name,
                dst_dp_rank=dp_j,
            )
            bcast_src = data_transfer_info.data_transfer_src_ranks[group_key]
            group = data_transfer_info.data_transfer_groups[group_key]
            dst_ranks = data_transfer_info.data_transfer_dst_ranks[group_key]

            buf_indices = [global_buffer_indices[_i] for _i in comm_slots]
            seqlens = [global_seqlens[_i] for _i in comm_slots]

            for dst_rank in dst_ranks:
                comm_plan.append(
                    DataTransferReceiverStep(
                        rank=dst_rank,
                        src=bcast_src,
                        dst_ranks=dst_ranks,
                        group=group,
                        key=k,
                        buf_indices=buf_indices,
                        seqlens=seqlens,
                    )
                )

            comm_plan.append(
                DataTransferSenderStep(
                    rank=bcast_src,
                    dst_ranks=dst_ranks,
                    group=group,
                    key=k,
                    buf_indices=buf_indices,
                    seqlens=seqlens,
                )
            )

    return comm_plan



from collections import defaultdict
from typing import Dict, List, Optional, Set, Tuple, Union
import asyncio
import collections
import copy
import dataclasses
import getpass
import itertools
import os
import re
import time
import uuid

import colorama
import deepspeed
import numpy as np
import torch
import torch.distributed

from realhf.api.core.config import ModelName
from realhf.api.core.model_api import ReaLModelConfig
from realhf.base import datapack, logging, namedarray, timeutil, topology
from realhf.base.asyncio_utils import (
    raise_asyncio_exception,
    setup_run_until_complete,
    teardown_run_util_complete,
)
from realhf.base.cluster import spec as cluster_spec
from realhf.base.constants import MODEL_SAVE_ROOT
from realhf.base.monitor import (
    caculuate_llama_forward_flops,
    calculate_llama_gen_flops,
    calculate_llama_train_flops,
)
from realhf.system.buffer import AsyncIOSequenceBuffer
import realhf.api.core.config as config_api
import realhf.api.core.data_api as data_api
import realhf.api.core.dfg as dfg
import realhf.api.core.model_api as model_api
import realhf.api.core.system_api as config_pkg
import realhf.base.recover as recover
import realhf.system.request_reply_stream as request_reply_stream
import realhf.system.worker_base as worker_base

logger = logging.getLogger("master worker", "system")
blogger = logging.getLogger("benchmark")


class ExperimentComplete(Exception):

    def __init__(self, message):
        disclaimer = (
            colorama.Fore.GREEN
            + "\033[1m"
            + "<This is not an error. It is just a way to stop the experiment.> "
        )
        super().__init__(
            disclaimer
            + colorama.Style.RESET_ALL
            + colorama.Fore.YELLOW
            + colorama.Style.BRIGHT
            + "\033[1m"
            + message
            + colorama.Style.RESET_ALL
        )


def request_all(
    stream: request_reply_stream.NameResolvingRequstClient,
    handlers: List[str],
    handle_type: str,
    datas: List[namedarray.NamedArray],
    verbose: bool = True,
):
    requests = [
        request_reply_stream.Payload(
            handler=handler,
            handle_name=handle_type,
            data=data,
        )
        for handler, data in zip(handlers, datas)
    ]
    if verbose:
        blogger.debug(
            f"master worker #request_all# *end* time ${time.time_ns()}$"
        )
    tik = time.perf_counter()

    # A protocol to ensure that any model worker execute jobs in the same order.
    [stream.post(r) for r in requests]
    [
        stream.poll(
            block=True, pattern=create_exact_match_pattern([p.syn_reply_id])
        )
        for p in requests
    ]
    [
        stream.post(
            request_reply_stream.Payload(
                handler=r.handler, handle_name="ack", request_id=r.ack_reply_id
            )
        )
        for r in requests
    ]
    t = time.perf_counter() - tik

    if verbose:
        blogger.debug(
            f'Request "{handle_type}" time in total: '
            f"{t:.4f}s, {t / len(requests):.4f}s per request"
        )
    return [r.request_id for r in requests]


def create_exact_match_pattern(
    string_list: List[Union[uuid.UUID, str]]
) -> re.Pattern:
    escaped_strings = [re.escape(str(s)) for s in string_list]
    pattern = f"({'|'.join(escaped_strings)})$"
    return re.compile(pattern)


async def _awaitable_response(
    stream: request_reply_stream.NameResolvingRequstClient,
    pattern: re.Pattern | None,
) -> request_reply_stream.Payload:
    while True:
        try:
            return stream.poll(pattern=pattern, block=False)
        except request_reply_stream.NoMessage:
            await asyncio.sleep(0.01)
            continue


async def gather_all_replies(
    stream: request_reply_stream.NameResolvingRequstClient,
    request_ids: List[str],
    verbose: bool = True,
) -> List:
    """Collect responses from multiple streams. Blocking method."""
    responses = await asyncio.gather(
        *[
            _awaitable_response(
                stream, pattern=create_exact_match_pattern([req_id])
            )
            for req_id in request_ids
        ]
    )
    if verbose:
        blogger.debug(
            f"master worker #gather_all_replies# *end* time ${time.time_ns()}$"
        )
    return responses


async def group_rpc_blocked(
    stream: request_reply_stream.NameResolvingRequstClient,
    handlers: List[Union[config_pkg.ModelShardID, str]],
    handle_type: str,
    datas: List[namedarray.NamedArray],
    verbose: bool = True,
) -> List[namedarray.NamedArray]:
    payloads = await gather_all_replies(
        stream,
        request_all(stream, handlers, handle_type, datas, verbose=verbose),
    )
    return [p.data for p in payloads]


def _request_parameter_sync(
    stream: request_reply_stream.NameResolvingRequstClient,
    msid2mwid: Dict[config_pkg.ModelShardID, int],
    from_model_name: ModelName,
    to_model_name: ModelName,
    from_topo: topology.PipeModelDataParallelTopology,
    to_topo: topology.PipeModelDataParallelTopology,
    to_model_config: ReaLModelConfig,
):

    model_name = from_model_name
    target = to_model_name
    # Prioritize handlers of `from_model`, then handlers of `to_model`.
    # As a result, if both `from_model` and `to_model` reside in a model worker,
    # the handler in the received request will be `from_model`. Layers will also built in `from_model`.
    # After that, we assign layers of the `from_model` to `to_model`.
    handlers = [
        config_pkg.ModelShardID.from_parallelism_rank(model_name, from_topo, j)
        for j in range(from_topo.world_size())
    ]
    all_handler_mwids = set([msid2mwid[h] for h in handlers])
    dst_handlers = [
        config_pkg.ModelShardID.from_parallelism_rank(target, to_topo, j)
        for j in range(to_topo.world_size())
    ]
    for h in dst_handlers:
        if msid2mwid[h] not in all_handler_mwids:
            handlers.append(h)
            all_handler_mwids.add(msid2mwid[h])

    ps_data = {
        "from_model_name": model_name,
        "to_model_name": target,
        "from_topo": from_topo,
        "to_topo": to_topo,
        "to_model_config": to_model_config,
    }
    payloads = [
        request_reply_stream.Payload(
            handler=h,
            handle_name="empty",
            pre_hooks=["param_realloc"],
            pre_hook_data=[ps_data],
        )
        for h in handlers
    ]
    request_ids = [stream.post(p) for p in payloads]
    [
        stream.poll(
            pattern=create_exact_match_pattern([p.syn_reply_id]), block=True
        )
        for p in payloads
    ]
    [
        stream.post(
            request_reply_stream.Payload(
                handler=p.handler, handle_name="ack", request_id=p.ack_reply_id
            )
        )
        for p in payloads
    ]
    [
        stream.poll(pattern=create_exact_match_pattern([req_id]), block=True)
        for req_id in request_ids
    ]


@dataclasses.dataclass
class InterfaceDataAmount:
    train_configs: List[ReaLModelConfig] = dataclasses.field(
        default_factory=list
    )
    train_bs: List[int] = dataclasses.field(default_factory=list)
    train_seqlens: List[List[int]] = dataclasses.field(default_factory=list)

    inf_configs: List[ReaLModelConfig] = dataclasses.field(default_factory=list)
    inf_bs: List[int] = dataclasses.field(default_factory=list)
    inf_seqlens: List[List[int]] = dataclasses.field(default_factory=list)

    gen_configs: List[ReaLModelConfig] = dataclasses.field(default_factory=list)
    gen_bs: List[int] = dataclasses.field(default_factory=list)
    prompt_lens: List[List[int]] = dataclasses.field(default_factory=list)
    gen_len: List[int] = dataclasses.field(default_factory=list)

    def clear(self):
        self.train_bs.clear()
        self.train_seqlens.clear()

        self.inf_bs.clear()
        self.inf_seqlens.clear()

        self.gen_bs.clear()
        self.prompt_lens.clear()
        self.gen_len.clear()

        self.train_configs.clear()
        self.inf_configs.clear()
        self.gen_configs.clear()


@dataclasses.dataclass
class RPCCorountineControl:
    ## Shared resources ##
    stop: asyncio.Event
    # for counting the number of finished training steps
    # one training step corresponds to traversal of the whole DFG
    train_count: asyncio.Queue
    # for loading data, save and eval model
    fetch_data_queue: asyncio.Queue
    data_spec_queue: asyncio.Queue
    eval_queue: asyncio.Queue
    save_queue: asyncio.Queue

    ## Per-coroutine resources ##
    # Used for counting the number of concurrent calls.
    can_do_rpc: Dict[str, asyncio.Semaphore]
    rpc_traversal: Dict[str, int]
    # for synchronizing req ids between req and reply coroutines
    request_queues: Dict[str, List[asyncio.Queue]]

    training_buffer_indices: Set[int] = dataclasses.field(default_factory=set)
    data_amount: InterfaceDataAmount = dataclasses.field(
        default_factory=InterfaceDataAmount
    )

    # recover information
    used_hash_vals_this_epoch: Set[int] = dataclasses.field(default_factory=set)
    is_recover_epoch: bool = False
    hash_vals_to_ignore_in_recover: Set[int] = dataclasses.field(
        default_factory=set
    )


def _attach_payloads_with_hooks(
    rpc: dfg.MFCDef,
    payloads: Dict[config_api.ModelShardID, request_reply_stream.Payload],
    mwids: List[int],
    msid2mwid: Dict[config_pkg.ModelShardID, int],
    model_configs: Dict[str, None | ReaLModelConfig],
    model_topos: Dict[str, topology.PipeModelDataParallelTopology],
    main_handlers: List[config_pkg.ModelShardID],
    hook_type: str,
) -> Tuple[
    Dict[config_api.ModelShardID, request_reply_stream.Payload], List[int]
]:
    assert hook_type in ["pre", "post"], hook_type

    main_mwids = set([msid2mwid[h] for h in main_handlers])
    for hook in getattr(rpc, f"{hook_type}_hooks"):
        if isinstance(hook, dfg.SyncParamHook):
            assert (hook.source is None) != (hook.target is None), hook
            if hook.source is None:
                src_topo = model_topos[rpc.model_name]
                dst_topo = model_topos[hook.target]
                dst_config = model_configs[hook.target]
                src_model_name, dst_model_name = rpc.model_name, hook.target
                other_model_name = hook.target
                other_topo = dst_topo
            else:
                src_topo = model_topos[hook.source]
                dst_topo = model_topos[rpc.model_name]
                dst_config = model_configs[rpc.model_name]
                src_model_name, dst_model_name = hook.source, rpc.model_name
                other_model_name = hook.source
                other_topo = src_topo

            ps_data = {
                "from_model_name": src_model_name,
                "to_model_name": dst_model_name,
                "from_topo": src_topo,
                "to_topo": dst_topo,
                "to_model_config": dst_config,
            }
            for h in main_handlers:
                getattr(payloads[h], f"{hook_type}_hooks").append(
                    "param_realloc"
                )
                getattr(payloads[h], f"{hook_type}_hook_data").append(ps_data)
            other_handlers = [
                config_api.ModelShardID.from_parallelism_rank(
                    other_model_name, other_topo, j
                )
                for j in range(other_topo.world_size())
            ]
            for h in other_handlers:
                if msid2mwid[h] not in mwids:
                    payloads[h] = request_reply_stream.Payload(
                        handler=h,
                        handle_name="empty",
                    )
                    setattr(
                        payloads[h], f"{hook_type}_hooks", ["param_realloc"]
                    )
                    setattr(payloads[h], f"{hook_type}_hook_data", [ps_data])
                    mwids.append(msid2mwid[h])
                elif msid2mwid[h] not in main_mwids:
                    hh = next(
                        hh for hh in payloads if msid2mwid[hh] == msid2mwid[h]
                    )
                    getattr(payloads[hh], f"{hook_type}_hooks").append(
                        "param_realloc"
                    )
                    getattr(payloads[hh], f"{hook_type}_hook_data").append(
                        ps_data
                    )

        elif isinstance(hook, dfg.OffloadHook):
            for h in main_handlers:
                getattr(payloads[h], f"{hook_type}_hooks").append("offload")
                getattr(payloads[h], f"{hook_type}_hook_data").append(
                    dict(model_name=h.model_name)
                )
        else:
            raise NotImplementedError(f"Unknown hook type: {hook}")
    return payloads, mwids


async def scatter_tensor_to_mws(
    rpc: dfg.MFCDef,
    stream: request_reply_stream.NameResolvingRequstClient,
    msid2mwid: Dict[config_pkg.ModelShardID, int],
    model_topos: Dict[str, topology.PipeModelDataParallelTopology],
    model_configs: Dict[str, None | ReaLModelConfig],
    producer_names: Dict[str, str],
    producer_name2producer_handlers: Dict[str, List[config_pkg.ModelShardID]],
    producer_mappings: Dict[str, Dict[str, List[int]]],
    target_mapping: Dict[str, List[int]],
    buffer_indices: List[int],
    seqlens: List[int],
    handlers: List[config_pkg.ModelShardID],
) -> List[uuid.UUID]:

    dt_data = {
        "keys": rpc.input_data,
        "target": rpc.model_name,
        "producer_names": producer_names,
        "producer_mappings": producer_mappings,
        "target_mapping": target_mapping,
        "handle_name": rpc.interface_type.value,
        "input_key_remap": rpc.input_key_remap,
        "output_key_remap": rpc.output_key_remap,
        "rpc_name": rpc.name,
        "buffer_indices": buffer_indices,
        "seqlens": seqlens,
    }

    payloads = {
        handler: request_reply_stream.Payload(
            handler=handler,
            handle_name=rpc.interface_type.value,
            pre_hooks=["data_transfer"],
            pre_hook_data=[dt_data],
        )
        for handler in handlers
    }
    mwids = [msid2mwid[h] for h in handlers]
    assert len(mwids) == len(set(mwids))

    for producer_name in producer_names.values():
        for h in producer_name2producer_handlers[producer_name]:
            if msid2mwid[h] not in mwids:
                payloads[h] = request_reply_stream.Payload(
                    handler=h,
                    handle_name="empty",
                    pre_hooks=["data_transfer"],
                    pre_hook_data=[dt_data],
                )
                mwids.append(msid2mwid[h])

    payloads, mwids = _attach_payloads_with_hooks(
        rpc,
        payloads,
        mwids,
        msid2mwid=msid2mwid,
        model_configs=model_configs,
        model_topos=model_topos,
        main_handlers=handlers,
        hook_type="pre",
    )
    payloads, mwids = _attach_payloads_with_hooks(
        rpc,
        payloads,
        mwids,
        msid2mwid=msid2mwid,
        model_configs=model_configs,
        model_topos=model_topos,
        main_handlers=handlers,
        hook_type="post",
    )
    req_ids = [stream.post(p) for h, p in payloads.items() if h in handlers]
    other_req_ids = [
        stream.post(p) for h, p in payloads.items() if h not in handlers
    ]
    await asyncio.gather(
        *[
            _awaitable_response(
                stream, pattern=create_exact_match_pattern([p.syn_reply_id])
            )
            for p in payloads.values()
        ]
    )
    [
        stream.post(
            request_reply_stream.Payload(
                handler=p.handler, handle_name="ack", request_id=p.ack_reply_id
            )
        )
        for p in payloads.values()
    ]
    return req_ids, other_req_ids


async def model_rpc_request_func(
    rpc: dfg.MFCDef,
    msid2mwid: Dict[config_pkg.ModelShardID, int],
    src_rpc_model_name: ModelName,
    stream: request_reply_stream.NameResolvingRequstClient,
    buffer: AsyncIOSequenceBuffer,
    data_owner: Dict[Tuple[int, str], Tuple[str, int]],
    model_topos: Dict[str, topology.PipeModelDataParallelTopology],
    model_configs: Dict[str, None | ReaLModelConfig],
    ctrl: RPCCorountineControl,
):
    topo = model_topos[rpc.model_name]
    handlers = [
        config_pkg.ModelShardID.from_parallelism_rank(rpc.model_name, topo, j)
        for j in range(topo.world_size())
    ]

    producer_names = {}  # data key -> model name
    for k in rpc.input_data:
        if k in rpc.data_producers:
            producer_names[k] = rpc.data_producers[k]
        else:
            producer_names[k] = src_rpc_model_name
    keys_to_send = defaultdict(list)  # model name -> List[keys] to send
    for k in producer_names:
        keys_to_send[producer_names[k]].append(k)

    # convert producer model name to ModelShardID
    producer_name2producer_handlers = {}
    for producer_name in keys_to_send:
        producer_name2producer_handlers[producer_name] = [
            config_pkg.ModelShardID.from_parallelism_rank(
                producer_name, model_topos[producer_name], j
            )
            for j in range(model_topos[producer_name].world_size())
        ]

    can_do_rpc = ctrl.can_do_rpc[rpc.name]
    request_queues = ctrl.request_queues[rpc.name]

    response_coroutine_idx = 0

    this_rpc_consumed_seqs = 0
    while not ctrl.stop.is_set():
        await can_do_rpc.acquire()

        # Ensure that parent RPCs will not be over-consumed.
        while any(
            this_rpc_consumed_seqs
            >= (ctrl.rpc_traversal[c.name] + 1) * c.max_n_seqs
            for c in rpc.children_rpcs
        ):
            await asyncio.sleep(0.1)

        sample = await buffer.get_batch_for_rpc(rpc)

        if rpc.is_src:
            ctrl.training_buffer_indices = ctrl.training_buffer_indices.union(
                sample.indices
            )
            ctrl.used_hash_vals_this_epoch = (
                ctrl.used_hash_vals_this_epoch.union(sample.hash_vals)
            )

        if rpc.interface_type == dfg.ModelInterfaceType.GENERATE:
            ctrl.data_amount.gen_configs.append(model_configs[rpc.model_name])
            ctrl.data_amount.gen_bs.append(len(sample.seqlens))
            ctrl.data_amount.gen_len.append(
                rpc.interface_impl.args["generation_config"]["min_new_tokens"]
            )
            ctrl.data_amount.prompt_lens.append(sample.seqlens)
        elif rpc.interface_type == dfg.ModelInterfaceType.TRAIN_STEP:
            ctrl.data_amount.train_configs.append(model_configs[rpc.model_name])
            ctrl.data_amount.train_bs.append(len(sample.seqlens))
            ctrl.data_amount.train_seqlens.append(sample.seqlens)
        elif rpc.interface_type == dfg.ModelInterfaceType.INFERENCE:
            ctrl.data_amount.inf_configs.append(model_configs[rpc.model_name])
            ctrl.data_amount.inf_bs.append(len(sample.seqlens))
            ctrl.data_amount.inf_seqlens.append(sample.seqlens)

        this_rpc_consumed_seqs += len(sample.seqlens)

        # logger.info(f"Model rpc {rpc.name} requesting.")
        dp_size = topo.get_dim("data")
        if rpc.balanced_dp:
            assert len(sample.seqlens) % dp_size == 0
            min_n_seqs_per_dp = len(sample.seqlens) // dp_size
        else:
            min_n_seqs_per_dp = 1
        partitions = datapack.min_abs_diff_partition(
            np.array(sample.seqlens, dtype=np.int32),
            dp_size,
            min_size=min_n_seqs_per_dp,
        )
        target_mapping = {
            i: list(range(v[0], v[1])) for i, v in enumerate(partitions)
        }

        # Set data owner of produced data by this RPC, such that downstream RPCs can know
        # whether to fetch these data.
        for dp_idx, (st, ed) in enumerate(partitions):
            for i in range(st, ed):
                for k in rpc.output_data:
                    if k in rpc.output_key_remap:
                        k = rpc.output_key_remap[k]
                    data_owner[sample.indices[i], k] = (rpc.model_name, dp_idx)

        # Get the data owner of this RPC's input data.
        producer_mappings: Dict[Tuple[str, str], List[Tuple[int, int]]] = {}
        for k in rpc.input_data:
            names, dp_indices = [], []
            for buf_idx in sample.indices:
                owner_name, dp_idx = data_owner[(buf_idx, k)]
                names.append(owner_name)
                dp_indices.append(dp_idx)
            assert len(set(names)) == 1
            producer_mapping = defaultdict(list)
            for i, dp_idx in enumerate(dp_indices):
                producer_mapping[dp_idx].append(i)
            producer_mapping = {
                k: sorted(v) for k, v in producer_mapping.items()
            }
            producer_mappings[names[0], k] = producer_mapping

        # send partitioned data to model workers
        req_ids, other_req_ids = await scatter_tensor_to_mws(
            rpc=rpc,
            stream=stream,
            msid2mwid=msid2mwid,
            model_topos=model_topos,
            model_configs=model_configs,
            producer_names=producer_names,
            producer_name2producer_handlers=producer_name2producer_handlers,
            producer_mappings=producer_mappings,
            target_mapping=target_mapping,
            buffer_indices=sample.indices,
            seqlens=sample.seqlens,
            handlers=handlers,
        )
        await request_queues[response_coroutine_idx].put(
            (req_ids, other_req_ids, time.perf_counter())
        )
        response_coroutine_idx = (response_coroutine_idx + 1) % len(
            request_queues
        )
        logger.info(f"Model rpc {rpc.name} requested.")


async def model_rpc_reply_func(
    corountine_idx: int,
    rpc: dfg.MFCDef,
    stream: request_reply_stream.NameResolvingRequstClient,
    buffer: AsyncIOSequenceBuffer,
    model_topos: Dict[str, topology.PipeModelDataParallelTopology],
    ctrl: RPCCorountineControl,
):
    topo = model_topos[rpc.model_name]
    dp_size = topo.get_dim("data")
    dp_head_indices = [
        topo.get_rank(data=i, pipe=topo.get_dim("pipe") - 1, model=0)
        for i in range(dp_size)
    ]
    dp_head_handlers = [
        config_pkg.ModelShardID.from_parallelism_rank(rpc.model_name, topo, i)
        for i in dp_head_indices
    ]

    request_queue = ctrl.request_queues[rpc.name][corountine_idx]
    can_do_rpc = ctrl.can_do_rpc[rpc.name]

    while not ctrl.stop.is_set():
        req_ids, other_req_ids, tik = await request_queue.get()

        # empty requests with parameter synchronization hooks may be issued
        await asyncio.gather(
            *[
                _awaitable_response(
                    stream, pattern=create_exact_match_pattern([req_id])
                )
                for req_id in other_req_ids
            ]
        )

        responses = await asyncio.gather(
            *[
                _awaitable_response(
                    stream, pattern=create_exact_match_pattern([req_id])
                )
                for req_id in req_ids
            ]
        )
        # logger.info(f"rpc {rpc.name} received responses {req_ids}")

        responses: List[request_reply_stream.Payload] = [
            responses[i] for i in dp_head_indices
        ]
        recv_tik = time.perf_counter()

        if (
            isinstance(responses[-1].data, dict)
            and responses[-1].data.get("seqlens") is not None
        ):
            res = []
            for k in responses[0].data["keys"]:
                if k in rpc.output_key_remap:
                    res.append(rpc.output_key_remap[k])
                else:
                    res.append(k)
        else:
            res = _gather_stat([response.data for response in responses])

        if rpc.log_return_value:
            logger.info(f"RPC name {rpc.name} returns {res}")

        can_do_rpc.release()

        ctrl.rpc_traversal[rpc.name] += 1

        if rpc.is_dst:
            await ctrl.train_count.put(1)
        else:
            buffer_indices = sum(
                [response.data["buffer_indices"] for response in responses], []
            )
            keys = res
            seqlens = sum(
                [response.data["seqlens"] for response in responses], []
            )
            await buffer.amend_batch(
                buffer_indices, [(keys, seqlen) for seqlen in seqlens]
            )

        logger.info(
            f"Model rpc {rpc.name} finished. Run time {time.perf_counter() - tik:.4f}s."
        )


async def load_data_func(
    src_rpc: dfg.MFCDef,
    src_rpc_dp_size: int,
    src_rpc_model_name: str,
    buffer: AsyncIOSequenceBuffer,
    data_owner: Dict[Tuple[int, str], Tuple[str, int]],
    stream: request_reply_stream.NameResolvingRequstClient,
    ctrl: RPCCorountineControl,
):
    while not ctrl.stop.is_set():
        await ctrl.fetch_data_queue.get()
        # fetch data from dataloader to fill the sequence buffer
        blogger.info(f"Filling data into the buffer in a new epoch.")
        fetch_data_start = time.perf_counter()

        loaded_data_batches = []

        is_final_batch = False
        while not is_final_batch:
            data_batches: List[data_api.DataBatchMeta] = (
                await group_rpc_blocked(
                    stream,
                    handlers=[f"__data{i}__" for i in range(src_rpc_dp_size)],
                    handle_type="fetch",
                    datas=[None for _ in range(src_rpc_dp_size)],
                    verbose=False,
                )
            )
            cur_epoch = data_batches[0].epoch

            loaded_data_batches += data_batches

            is_final_batch = data_batches[0].is_final_batch
            assert all(is_final_batch == x.is_final_batch for x in data_batches)

        # PyTorch dataloader will shuffle data for us.
        all_data = []
        for x in loaded_data_batches:
            all_data += data_api.unpack_data_batch(x)

        steps_per_epoch = len(all_data) // src_rpc.max_n_seqs
        avg_tokens_per_batch = (
            sum(x.seqlens[0] for x in all_data) / steps_per_epoch
        )
        logger.info(
            f"Training epoch {cur_epoch + 1} approximately has {steps_per_epoch} steps. "
            f"Each batch has {avg_tokens_per_batch:.2f} tokens in average."
        )
        await ctrl.data_spec_queue.put((steps_per_epoch, avg_tokens_per_batch))

        dp_data_batches = [0 for _ in range(src_rpc_dp_size)]
        for x in all_data:
            dp_data_batches[x.dp_rank] += 1

        assert all(len(x.seqlens) == 1 for x in all_data)
        seqlens = np.array([x.seqlens[0] for x in all_data], dtype=np.int32)

        # NOTE: The reordered indices prioritize longer sequences for detecting OOM errors early.
        reorder_indices, _ = datapack.reorder_to_balanced_batches(
            seqlens, src_rpc.max_n_seqs
        )
        recover_indices = np.argsort(reorder_indices)
        all_data: List[data_api.DataBatchMeta] = [
            all_data[i] for i in reorder_indices
        ]

        keys = all_data[0].keys
        batch = [(keys, x.seqlens[0], x.hash_vals[0]) for x in all_data]
        if ctrl.is_recover_epoch:
            batch = list(
                filter(
                    lambda x: x[2] not in ctrl.hash_vals_to_ignore_in_recover,
                    batch,
                )
            )
        buffer_indices = await buffer.put_batch(batch)

        all_data = [all_data[i] for i in recover_indices]
        buffer_indices = [buffer_indices[i] for i in recover_indices]
        assert len(buffer_indices) == len(all_data)

        for k in keys:
            for buf_idx, x in zip(buffer_indices, all_data):
                data_owner[(buf_idx, k)] = (src_rpc_model_name, x.dp_rank)

        store_buffer_indices = [[] for _ in range(src_rpc_dp_size)]
        for buf_idx, x in zip(buffer_indices, all_data):
            store_buffer_indices[x.dp_rank].append(buf_idx)

        for x, y in zip(store_buffer_indices, dp_data_batches):
            assert len(x) == y, (len(x), y)

        await group_rpc_blocked(
            stream,
            handlers=[f"__data{i}__" for i in range(src_rpc_dp_size)],
            handle_type="store",
            datas=store_buffer_indices,
            verbose=False,
        )

        async with buffer.lock:
            buffer.lock.notify(buffer.n_rpcs)

        blogger.info(
            f"Filling data finished. Time consumption: {time.perf_counter() - fetch_data_start:.3f}s."
        )


def _gather_stat(src: List[Dict]) -> Dict:
    cnt, stats = {}, {}
    for reply in src:
        for k, v in reply.items():
            cnt[k] = cnt.get(k, 0) + 1
            stats[k] = stats.get(k, 0) + v
    res = {
        k: v / cnt
        for k, v, cnt in zip(stats.keys(), stats.values(), cnt.values())
    }
    for k, c in cnt.items():
        if c != len(src):
            logger.warning(
                f"Gathered `{k}` is not present in every returned stats."
            )
    for k, v in res.items():
        if any(abs(v - x.get(k, None)) > 1e-4 for x in src):
            logger.warning(
                f"Gathered `{k}` is not all-reduced before returning: ({[x.get(k, None) for x in src]}, {v})."
            )
    return res


async def model_eval_thread_func(
    stream: request_reply_stream.NameResolvingRequstClient,
    handlers: List[config_pkg.ModelShardID],
    eval_queue: asyncio.Queue,
    stop_ctl: asyncio.Event,
):
    while not stop_ctl.is_set():
        epoch, epoch_step = await eval_queue.get()
        eval_stats = await group_rpc_blocked(
            stream, handlers, "evaluate", [None for _ in handlers]
        )
        eval_stats = _gather_stat(list(filter(lambda x: bool(x), eval_stats)))
        logger.info(
            f"Evaluation results at epoch {epoch + 1} step {epoch_step + 1}: {eval_stats}"
        )


async def model_save_thread_func(
    stream: request_reply_stream.NameResolvingRequstClient,
    handlers: List[config_pkg.ModelShardID],
    model_save_root: str,
    save_queue: asyncio.Queue,
    stop_ctl: asyncio.Event,
):
    while not stop_ctl.is_set():
        epoch, epoch_step, global_step = await save_queue.get()
        model_save_dirs = [
            os.path.join(
                model_save_root,
                s.model_name.role,
                f"epoch{epoch}epochstep{epoch_step}globalstep{global_step}",
            )
            for s in handlers
        ]
        await group_rpc_blocked(stream, handlers, "save", model_save_dirs)
        logger.info(f"Save models at epoch {epoch} step {epoch_step}.")


class MasterWorker(worker_base.Worker):
    os.makedirs(MODEL_SAVE_ROOT, exist_ok=True)

    def _configure(self, config: config_pkg.MasterWorker):
        self.config = config

        self.__model_topos: Dict[
            ModelName, topology.PipeModelDataParallelTopology
        ] = config.model_topos

        # Build execution graph and initialize concurrency utilities.
        self.__model_rpcs, _ = dfg.build_graph(config.model_rpcs)
        for rpc in self.__model_rpcs:
            _dp_size = self.__model_topos[rpc.model_name].get_dim("data")
            _pp_size = self.__model_topos[rpc.model_name].get_dim("pipe")
            if rpc.min_n_seqs < _dp_size * _pp_size:
                logger.warning(
                    f"The batch size of RPC `{rpc.name}` in terms of #seqs is smaller than "
                    f"dp_size * pp_size ({_dp_size}*{_pp_size}). Forcely enlarge the batch size "
                    f"to {_dp_size * _pp_size} (dp_size * pp_size). (original: {rpc.min_n_seqs})"
                )
                rpc.min_n_seqs_per_dp = 1
                rpc.min_n_seqs = _dp_size * _pp_size

        self.__mwid2msids = defaultdict(list)
        for msid, mwid in self.config.msid2mwid.items():
            self.__mwid2msids[mwid].append(msid)

        self.__rpc_srcs = list(
            filter(lambda rpc: rpc.is_src, self.__model_rpcs)
        )
        self.__rpc_dsts = list(
            filter(lambda rpc: rpc.is_dst, self.__model_rpcs)
        )
        self.__n_rpc_srcs = len(self.__rpc_srcs)
        self.__n_rpc_dsts = len(self.__rpc_dsts)

        # Save and eval control.
        self.__total_train_epochs = config.exp_ctrl.total_train_epochs
        self.__save_ctl = timeutil.EpochStepTimeFreqCtl(
            freq_epoch=config.exp_ctrl.save_frequency_epochs,
            freq_step=config.exp_ctrl.save_frequency_steps,
            freq_sec=config.exp_ctrl.save_frequency_seconds,
        )
        self.__eval_ctl = timeutil.EpochStepTimeFreqCtl(
            freq_epoch=config.exp_ctrl.eval_frequency_epochs,
            freq_step=config.exp_ctrl.eval_frequency_steps,
            freq_sec=config.exp_ctrl.eval_frequency_seconds,
        )

        self.MODEL_SAVE_ROOT = os.path.join(
            MODEL_SAVE_ROOT,
            config.worker_info.experiment_name,
            config.worker_info.trial_name,
        )
        os.makedirs(self.MODEL_SAVE_ROOT, exist_ok=True)

        self.__initialized = False
        self.__recover_run = os.environ.get("RECOVER_RUN", "0") == "1"
        self.__recover_info = (
            recover.load_recover_info() if self.__recover_run else None
        )
        self.__recover_first_epoch_done = False

        self._epoch = 0
        self._epoch_step = self._global_step = 0
        if self.__recover_run:
            self._epoch = self.__recover_info.recover_start.epoch
            self._epoch_step = self.__recover_info.recover_start.epoch_step
            self._start_global_step = self._global_step = (
                self.__recover_info.recover_start.global_step
            )
            logger.info(
                f"Recovering from previous run: "
                f"{(self._epoch, self._epoch_step, self._global_step)}"
            )

        self.__this_step_info = recover.StepInfo(
            epoch=0,
            epoch_step=0,
            global_step=0,
        )
        self.__last_step_info = None

        # for benchmark
        self.e2e_time_history = []
        self.level_time_history = defaultdict(list)
        self.__benchmark_steps = config.exp_ctrl.benchmark_steps

        return config.worker_info

    def __lazy_init(self):
        # Set up streams.
        handler_routing = copy.deepcopy(self.config.msid2mwid)
        src_rpc = self.__rpc_srcs[0]
        src_rpc_topo = self.config.model_topos[src_rpc.model_name]
        src_rpc_dp_size = src_rpc_topo.get_dim("data")
        src_rpc_pp_size = src_rpc_topo.get_dim("pipe")
        for i in range(src_rpc_dp_size):
            rank = src_rpc_topo.get_rank(
                data=i, pipe=src_rpc_pp_size - 1, model=0
            )
            handler_routing[f"__data{i}__"] = self.config.msid2mwid[
                config_pkg.ModelShardID.from_parallelism_rank(
                    model_name=src_rpc.model_name,
                    topo=src_rpc_topo,
                    parallelism_rank=rank,
                )
            ]
        self.__stream = request_reply_stream.make_master_stream(
            self.config.worker_info,
            n_subscribers=self.config.n_model_workers,
            handler_routing=handler_routing,
        )
        self.__stream: request_reply_stream.NameResolvingRequstClient

        src_rpc = [rpc for rpc in self.config.model_rpcs if rpc.is_src][0]
        src_rpc_model_name = src_rpc.model_name
        src_rpc_dp_size = self.config.model_topos[src_rpc.model_name].get_dim(
            "data"
        )

        # Request training specification from data workers.
        total_n_seqs = 0
        for i in range(src_rpc_dp_size):
            p = request_reply_stream.Payload(
                handler=f"__data{i}__",
                handle_name="spec",
            )
            self.__stream.post(p)
            self.__stream.poll(
                block=True, pattern=create_exact_match_pattern([p.syn_reply_id])
            )
            self.__stream.post(
                request_reply_stream.Payload(
                    handler=f"__data{i}__",
                    handle_name="ack",
                    request_id=p.ack_reply_id,
                )
            )
            n_seqs_per_dataset_shard: int = self.__stream.poll(
                block=True, pattern=create_exact_match_pattern([p.request_id])
            ).data
            total_n_seqs += n_seqs_per_dataset_shard

        event_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(event_loop)

        # Build some data required for subsequent model function calls.
        self.__all_model_handlers: List[config_pkg.ModelShardID] = []
        self.__dp0_model_handlers: List[config_pkg.ModelShardID] = []
        for model_name, topo in self.config.model_topos.items():
            num_dp = topo.get_dim("data")
            self.__all_model_handlers += [
                config_pkg.ModelShardID.from_parallelism_rank(
                    model_name, topo, j
                )
                for j in range(topo.world_size())
            ]
            self.__dp0_model_handlers += [
                config_pkg.ModelShardID.from_parallelism_rank(
                    model_name, topo, j
                )
                for j in topo.filter_match(data=0)
            ]

        # Request model configs from model workers.
        # Return None if the model is not a ReaLModel.
        self.__model_configs: Dict[ModelName, None | ReaLModelConfig] = {}
        for model_name in self.config.model_topos:
            p = request_reply_stream.Payload(
                handler=config_pkg.ModelShardID.from_parallelism_rank(
                    model_name, topo, 0
                ),
                handle_name="model_config",
            )
            self.__stream.post(p)
            self.__stream.poll(
                block=True, pattern=create_exact_match_pattern([p.syn_reply_id])
            )
            self.__stream.post(
                request_reply_stream.Payload(
                    handler=p.handler,
                    handle_name="ack",
                    request_id=p.ack_reply_id,
                )
            )
            self.__model_configs[model_name] = self.__stream.poll(
                pattern=create_exact_match_pattern([p.request_id]), block=True
            ).data

        # Initialize model backends.
        # For models with the same role, they share the same model parameters.
        # Therefore, we must call reallocate parameters from A to B
        # before we send requests to initialize B.
        _param_senders = [v[0] for v in self.config.sync_param_pairs]
        _param_recevers = [v[1] for v in self.config.sync_param_pairs]

        # The parameters are by default held by the trainable model.
        # If all replicas are not trainable, the parameters are held in replica 0.
        _model_is_trainable = collections.defaultdict(list)
        for rpc in self.__model_rpcs:
            _model_is_trainable[rpc.model_name].append(
                rpc.interface_type == dfg.ModelInterfaceType.TRAIN_STEP
            )

        _model_is_trainable = {
            model_name: any(values)
            for model_name, values in _model_is_trainable.items()
        }

        _roles = set([rpc.model_name.role for rpc in self.__model_rpcs])
        _role_cnt = {
            role: len(
                set(
                    [
                        rpc.model_name
                        for rpc in self.__model_rpcs
                        if rpc.model_name.role == role
                    ]
                )
            )
            for role in _roles
        }
        _reordered_model_names = []
        for role in _roles:
            if _role_cnt[role] == 1:
                _reordered_model_names.append(ModelName(role, 0))
                continue
            _indices = list(range(_role_cnt[role]))
            _trainable_this_role = [
                _model_is_trainable[ModelName(role, i)]
                for i in range(_role_cnt[role])
            ]
            if any(_trainable_this_role):
                assert sum(_trainable_this_role) == 1
                _trainable_idx = _trainable_this_role.index(True)
                _reordered_model_names.append(ModelName(role, _trainable_idx))
                _indices.remove(_trainable_idx)
            for i in _indices:
                _reordered_model_names.append(ModelName(role, i))

        # Send initialization requests.
        self.logger.info(
            f"Initialize model backends with order: {_reordered_model_names}."
        )
        _initialized_roles = []
        for model_name in _reordered_model_names:
            topo = self.config.model_topos[model_name]
            # Build FinetuneSpec, which is required to initialize backends.
            _handlers = [
                config_pkg.ModelShardID.from_parallelism_rank(
                    model_name, topo, j
                )
                for j in range(topo.world_size())
            ]
            train_rpcs = list(
                filter(
                    lambda rpc: rpc.model_name == model_name
                    and rpc.interface_type == dfg.ModelInterfaceType.TRAIN_STEP,
                    self.__model_rpcs,
                )
            )
            if len(train_rpcs) > 0:
                assert len(train_rpcs) == 1
                steps_per_epoch = (
                    total_n_seqs * self.config.exp_ctrl.total_train_epochs
                    + train_rpcs[0].max_n_seqs
                    - 1
                ) // train_rpcs[0].max_n_seqs
                ft_spec = model_api.FinetuneSpec(
                    total_train_epochs=self.config.exp_ctrl.total_train_epochs,
                    total_train_steps=steps_per_epoch
                    * self.config.exp_ctrl.total_train_epochs,
                    steps_per_epoch=steps_per_epoch,
                )
            else:
                ft_spec = model_api.FinetuneSpec(
                    total_train_epochs=self.config.exp_ctrl.total_train_epochs,
                    total_train_steps=-1,
                    steps_per_epoch=-1,
                )
            model_ft_specs = [ft_spec] * topo.world_size()

            # Reallocate parameters if necessary.
            if model_name.role in _initialized_roles:
                assert model_name in _param_recevers
                _param_realloc_src = _param_senders[
                    _param_recevers.index(model_name)
                ]
                _request_parameter_sync(
                    stream=self.__stream,
                    msid2mwid=self.config.msid2mwid,
                    from_model_name=_param_realloc_src,
                    to_model_name=model_name,
                    from_topo=self.config.model_topos[_param_realloc_src],
                    to_topo=self.config.model_topos[model_name],
                    to_model_config=self.__model_configs[model_name],
                )

            _task = event_loop.create_task(
                group_rpc_blocked(
                    self.__stream,
                    handlers=_handlers,
                    handle_type="initialize",
                    datas=model_ft_specs,
                )
            )
            event_loop.run_until_complete(asyncio.gather(_task))[0]

            # Reallocate parameters back.
            if model_name.role in _initialized_roles:
                # Reversely sync parameters
                _request_parameter_sync(
                    stream=self.__stream,
                    msid2mwid=self.config.msid2mwid,
                    from_model_name=model_name,
                    to_model_name=_param_realloc_src,
                    to_topo=self.config.model_topos[_param_realloc_src],
                    from_topo=self.config.model_topos[model_name],
                    to_model_config=self.__model_configs[_param_realloc_src],
                )

            _initialized_roles.append(model_name.role)

        logger.info("Initializations of models and backends complete.")

        # Create corountine control objects for running the dataflow graph.
        self.__rpc_ctrl = RPCCorountineControl(
            stop=asyncio.Event(),
            train_count=asyncio.Queue(maxsize=len(self.__rpc_dsts)),
            fetch_data_queue=asyncio.Queue(1),
            data_spec_queue=asyncio.Queue(1),
            eval_queue=asyncio.Queue(1),
            save_queue=asyncio.Queue(1),
            rpc_traversal={rpc.name: 0 for rpc in self.__model_rpcs},
            can_do_rpc={
                rpc.name: asyncio.Semaphore(rpc.max_concurrent_calls)
                for rpc in self.__model_rpcs
            },
            request_queues={
                rpc.name: [
                    asyncio.Queue(1) for _ in range(rpc.max_concurrent_calls)
                ]
                for rpc in self.__model_rpcs
            },
            is_recover_epoch=self.__recover_run,
            used_hash_vals_this_epoch=(
                self.__recover_info.hash_vals_to_ignore
                if self.__recover_run
                else set()
            ),
            hash_vals_to_ignore_in_recover=(
                self.__recover_info.hash_vals_to_ignore
                if self.__recover_run
                else set()
            ),
        )

        self.__fetch_master_ctl = asyncio.Queue(1)

        # NOTE: We don't set a configurable maximum buffer size here because we want to keep all data in the buffer.
        # We assume that the dataset size is at most 1M. We have warnings if the buffer is 95% full.
        self.__seqbuffer = AsyncIOSequenceBuffer(
            self.__model_rpcs,
            max_size=int(1e6),
            fetch_ctl=self.__rpc_ctrl.fetch_data_queue,
            fetch_master_ctl=self.__fetch_master_ctl,
        )

        self.__data_owner: Dict[Tuple[int, str], Tuple[str, int]] = {}

        logger.info(f"Creating asyncio coroutines...")

        # Create coroutines for model RPCs.
        coroutine_tasks = []
        for rpc in self.__model_rpcs:
            request_task = event_loop.create_task(
                model_rpc_request_func(
                    rpc=rpc,
                    msid2mwid=self.config.msid2mwid,
                    src_rpc_model_name=src_rpc_model_name,
                    data_owner=self.__data_owner,
                    stream=self.__stream,
                    buffer=self.__seqbuffer,
                    model_topos=self.__model_topos,
                    model_configs=self.__model_configs,
                    ctrl=self.__rpc_ctrl,
                )
            )
            reply_tasks = []
            for j in range(rpc.max_concurrent_calls):
                _reply_task = event_loop.create_task(
                    model_rpc_reply_func(
                        corountine_idx=j,
                        rpc=rpc,
                        stream=self.__stream,
                        buffer=self.__seqbuffer,
                        model_topos=self.__model_topos,
                        ctrl=self.__rpc_ctrl,
                    )
                )
                reply_tasks.append(_reply_task)
            coroutine_tasks += [request_task] + reply_tasks

        # Append some utilization coroutines to handle data loading, saving, and evaluation.
        load_data_task = event_loop.create_task(
            load_data_func(
                src_rpc=src_rpc,
                src_rpc_dp_size=src_rpc_dp_size,
                src_rpc_model_name=src_rpc_model_name,
                data_owner=self.__data_owner,
                buffer=self.__seqbuffer,
                stream=self.__stream,
                ctrl=self.__rpc_ctrl,
            )
        )
        eval_task = event_loop.create_task(
            model_eval_thread_func(
                stream=self.__stream,
                handlers=self.__all_model_handlers,
                eval_queue=self.__rpc_ctrl.eval_queue,
                stop_ctl=self.__rpc_ctrl.stop,
            )
        )
        save_task = event_loop.create_task(
            model_save_thread_func(
                stream=self.__stream,
                handlers=self.__all_model_handlers,
                model_save_root=self.MODEL_SAVE_ROOT,
                save_queue=self.__rpc_ctrl.save_queue,
                stop_ctl=self.__rpc_ctrl.stop,
            )
        )
        coroutine_tasks += [load_data_task, eval_task, save_task]

        # Set up a run context of EventLoop.run_util_complete, baiscally copy-paste from cpython.
        # With this context, we can call the non-block EventLoop._run_once (similar to worker._poll).
        self.__asyncio_ctx = setup_run_until_complete(
            event_loop, asyncio.gather(*coroutine_tasks)
        )

        logger.info(f"Coroutines created. The master worker is ready to run.")

        self.__initialized = True
        self._train_start_time = time.perf_counter()

        self.__clear_data_cache_reqids = None

        self.__cur_steps_per_epoch = None
        self.__cur_avg_tokens_per_batch = None

    def _poll(self):
        if not self.__initialized:
            self.__lazy_init()

        # Main execution steps. The graph runs under-the-hood in RPC & stream threads.
        # Wait for the finish of the traversal of the execution graph.
        execution_start = time.perf_counter()
        logger.info(
            "Master worker is waiting for the finish of the execution graph."
        )
        _rpc_dst_cnt = 0
        while _rpc_dst_cnt < self.__n_rpc_dsts:
            try:
                self.__rpc_ctrl.train_count.get_nowait()
                _rpc_dst_cnt += 1
                continue
            except asyncio.QueueEmpty:
                pass
            try:
                # Similar to worker._poll. Run multiple times until a train step is finished.
                self.__asyncio_ctx.loop._run_once()
                # NOTE: The following line will propagate errors in corountines back to the main thread.
                # It raises asyncio.exceptions.InvalidStateError if the result is not ready.
                # (In our use cases, the result will never be ready because corountines run while-loops.)
                # We just ignore this error and continue running.
                self.__asyncio_ctx.future.result()
            except asyncio.exceptions.InvalidStateError:
                # Catch the exception when future.result() is not ready.
                pass
            except KeyboardInterrupt as e:
                raise_asyncio_exception(self.__asyncio_ctx, raise_error=False)
                raise e
            except:
                raise_asyncio_exception(self.__asyncio_ctx)
        logger.info("Execution finished!")

        try:
            self.__fetch_master_ctl.get_nowait()
            is_new_epoch = True
        except asyncio.QueueEmpty:
            is_new_epoch = False

        try:
            (self.__cur_steps_per_epoch, self.__cur_avg_tokens_per_batch) = (
                self.__rpc_ctrl.data_spec_queue.get_nowait()
            )
        except asyncio.QueueEmpty:
            pass

        should_eval = self.__eval_ctl.check(epochs=int(is_new_epoch), steps=1)
        should_save = self.__save_ctl.check(epochs=int(is_new_epoch), steps=1)

        if is_new_epoch:
            # do not update epoch counter when recover for the first step
            if self.__recover_run and not self.__recover_first_epoch_done:
                self.__recover_first_epoch_done = True
            else:
                self._epoch += 1
                self._epoch_step = 0
                self.__rpc_ctrl.used_hash_vals_this_epoch = set()

        self._epoch_step += 1
        self._global_step += 1

        if should_eval:
            self.__rpc_ctrl.eval_queue.put_nowait(
                (self._epoch, self._epoch_step)
            )
        if should_save:
            self.__rpc_ctrl.save_queue.put_nowait(
                (self._epoch, self._epoch_step, self._global_step)
            )

        # update step info
        self.__last_step_info = self.__this_step_info
        self.__this_step_info = recover.StepInfo(
            epoch=self._epoch,
            epoch_step=self._epoch_step,
            global_step=self._global_step,
        )

        if is_new_epoch:
            if self._epoch > self.__total_train_epochs:
                self.experiment_complete_exit(f"Training completes! Yeah!!!")

        total_time_consumption = time.perf_counter() - self._train_start_time
        time_per_step = total_time_consumption / (self._global_step + 1)
        e2e_time = time.perf_counter() - execution_start
        self.e2e_time_history.append(e2e_time)

        # calculate flops
        #########################################
        if not all(
            isinstance(v, ReaLModelConfig)
            for v in self.__model_configs.values()
        ):
            logger.warning(
                f"Not all models are ReaLModels. Unable to calculate FLOP/s."
            )
            flops = None
        else:
            flops = 0
            for train_bs, train_seqlens, real_config in zip(
                self.__rpc_ctrl.data_amount.train_bs,
                self.__rpc_ctrl.data_amount.train_seqlens,
                self.__rpc_ctrl.data_amount.train_configs,
            ):
                flops += calculate_llama_train_flops(
                    checkpoint_activations_factor=4,
                    batch_size=train_bs,
                    seqlens=train_seqlens,
                    num_layers=real_config.n_layers,
                    hidden_size=real_config.hidden_dim,
                    intermediate_size=real_config.intermediate_dim,
                    vocab_size=real_config.vocab_size,
                )
            for inf_bs, inf_seqlens, real_config in zip(
                self.__rpc_ctrl.data_amount.inf_bs,
                self.__rpc_ctrl.data_amount.inf_seqlens,
                self.__rpc_ctrl.data_amount.inf_configs,
            ):
                flops += caculuate_llama_forward_flops(
                    batch_size=inf_bs,
                    seqlens=inf_seqlens,
                    num_layers=real_config.n_layers,
                    hidden_size=real_config.hidden_dim,
                    intermediate_size=real_config.intermediate_dim,
                    vocab_size=real_config.vocab_size,
                )
            for gen_bs, prompt_lens, gen_len, real_config in zip(
                self.__rpc_ctrl.data_amount.gen_bs,
                self.__rpc_ctrl.data_amount.prompt_lens,
                self.__rpc_ctrl.data_amount.gen_len,
                self.__rpc_ctrl.data_amount.gen_configs,
            ):
                flops += calculate_llama_gen_flops(
                    batch_size=gen_bs,
                    prompt_lens=prompt_lens,
                    gen_len=gen_len,
                    num_layers=real_config.n_layers,
                    hidden_size=real_config.hidden_dim,
                    intermediate_size=real_config.intermediate_dim,
                    vocab_size=real_config.vocab_size,
                )
            tflops = flops / (e2e_time * (10**12))
            tflops_per_gpu = flops / (
                e2e_time * self.config.n_model_workers * (10**12)
            )
        self.__rpc_ctrl.data_amount.clear()
        #########################################

        # Logging.
        s = f"Epoch {self._epoch}/{self.config.exp_ctrl.total_train_epochs} "
        if self.__cur_steps_per_epoch is not None:
            s += f"step {self._epoch_step}/{self.__cur_steps_per_epoch} "
        else:
            s += f"step {self._epoch_step} "
        s += f"(global step {self._global_step}) finishes. "
        if self.__cur_avg_tokens_per_batch is not None:
            s += f"Average #tokens per batch is {self.__cur_avg_tokens_per_batch:.0f}. "
        s += f"#End to end# execution time: *{e2e_time:.3f}*s. "
        s += f"Total time consumption: {total_time_consumption:.3f}s. "
        if (
            self.__cur_steps_per_epoch is not None
            and len(self.e2e_time_history) > 2
        ):
            remaining_steps = self.__cur_steps_per_epoch - self._epoch_step
            remaining_epochs = self.__total_train_epochs - self._epoch
            avg_t = np.mean(self.e2e_time_history[2:])
            remain_t = avg_t * remaining_steps
            remain_t += avg_t * self.__cur_steps_per_epoch * remaining_epochs
            s += f"Estimated remaining time: {remain_t:.3f}s. "
        if flops is not None:
            s += f"TFLOP/s per GPU: {tflops_per_gpu:.2f}, total TFLOP/s: {tflops:.2f}."
        logger.info(s)

        if (
            self.__benchmark_steps is not None
            and self._global_step >= self.__benchmark_steps
        ):
            logger.info(
                f"Finished benchmark {self.__benchmark_steps}. Total time consumption {total_time_consumption:.3f}"
            )
            logger.info(
                f"avg #e2e# time *{np.mean(self.e2e_time_history):.3f}*"
            )
            for i, level_time_history in self.level_time_history.items():
                logger.info(
                    f"avg #level{i+1}# time *{np.mean(level_time_history):.3f}*"
                )
            self.experiment_complete_exit(f"Benchmark completes! Yeah!!!")

        if self.__clear_data_cache_reqids is not None:
            [
                self.__stream.poll(
                    block=True, pattern=create_exact_match_pattern([reqid])
                )
                for reqid in self.__clear_data_cache_reqids
            ]
        self.__clear_data_cache_reqids = request_all(
            self.__stream,
            [vs[0] for vs in self.__mwid2msids.values()],
            "clear_data_cache",
            [
                self.__rpc_ctrl.training_buffer_indices
                for _ in self.__all_model_handlers
            ],
        )
        self.__rpc_ctrl.training_buffer_indices.clear()

        return worker_base.PollResult(sample_count=1, batch_count=1)

    def experiment_complete_exit(self, msg: str):
        self.__rpc_ctrl.stop.set()
        self.__asyncio_ctx.loop.stop()
        try:
            teardown_run_util_complete(self.__asyncio_ctx)
        except RuntimeError as e:
            logger.info(
                colorama.Style.RESET_ALL
                + colorama.Fore.YELLOW
                + colorama.Style.BRIGHT
                + "\033[1m"
                + msg
                + colorama.Style.RESET_ALL
            )
            self.exit()
            # raise ExperimentComplete(msg) from e

    def __recover_save(self):
        # save step info for recover
        recover_info = recover.RecoverInfo(
            # recover_start=self.__last_step_info,
            recover_start=self.__this_step_info,
            last_step_info=self.__last_step_info,
            hash_vals_to_ignore=self.__rpc_ctrl.used_hash_vals_this_epoch,
        )
        import pprint

        logger.info("dumped recover info to file")
        pprint.pprint(recover_info.recover_start)
        pprint.pprint(recover_info.last_step_info)
        pprint.pprint(len(recover_info.hash_vals_to_ignore))
        recover.dump_recover_info(recover_info)

    def _exit_hook(self, exit_status: worker_base.WorkerServerStatus):
        logger.info(f"Master worker exits with {exit_status}.")
        if os.environ["SAVE_RECOVER_STATES"] == "0":
            return
        if exit_status == worker_base.WorkerServerStatus.ERROR:
            try:
                sleep_time = 600
                current_sleep_time = 0
                while current_sleep_time < sleep_time:
                    logger.info(
                        f"ERROR exit, waited {current_sleep_time} s for interruption ..."
                    )
                    time.sleep(10)
                    current_sleep_time += 10
            except KeyboardInterrupt:
                logger.info("Received SIGINT, starting recover save")

        self.__recover_save()
